from ocr.engine import OCRe  # 引擎单例
from utils.asset import Asset  # 资源
from utils.config import Config
from ui.widget import Widget  # 控件

import tkinter as tk
from tkinter import ttk
import tkinter.messagebox
from windnd import hook_dropfiles  # 文件拖拽
from PIL import Image, ImageTk
import os


class IgnoreAreaWin:
    def __init__(self, closeSendData=None, defaultPath=""):
        self.closeSendData = closeSendData  # 向父窗口发送数据的接口
        self.balloon = Config.main.balloon  # 气泡框
        self.cW = 960  # 画板尺寸
        self.cH = 540
        self.tran = 2  # 绘制偏移量
        self.areaColor = ["red", "green",  # 各个矩形区域的标志颜色
                          "darkorange", "white"]

        # def initWin():  # 初始化主窗口
        # 主窗口
        self.win = tk.Toplevel()
        self.win.protocol("WM_DELETE_WINDOW", self.onClose)
        self.win.title("选择区域")
        self.win.resizable(False, False)  # 禁止窗口拉伸
        # 变量初始化
        self.imgSize = (-1, -1)  # 图像尺寸，初次加载时生效。
        self.imgScale = -1  # 图片缩放比例
        self.imgSizeText = tk.StringVar(value="未设定")
        self.area = [[], [], []]  # 存储各个矩形区域的列表
        self.areaHistory = []  # 绘图历史，撤销用
        self.areaTextRec = []  # 文字区域提示框
        self.areaType = -1  # 当前绘图模式
        self.areaTypeIndex = [-1, -1, -1]  # 当前绘制的矩形的序号
        self.lastPath = ''  # 上一次导入的图片的路径
        # 图标
        self.win.iconphoto(False, Asset.getImgTK('umiocr24'))  # 设置窗口图标
        # initWin()

        # def initPanel():  # 初始化面板
        tk.Frame(self.win, height=10).pack(side='top')
        ctrlFrame = tk.Frame(self.win)
        ctrlFrame.pack(side='top', fill='y')
        ctrlF0 = tk.Frame(ctrlFrame)
        ctrlF0.pack(side='left')
        tk.Label(ctrlF0, text="【图像分辨率】").pack()
        tk.Label(ctrlF0, textvariable=self.imgSizeText).pack()
        tk.Label(ctrlF0, text="不符合该分辨率的图片\n忽略区域设置不会生效",
                 fg='gray', wraplength=120).pack()
        tk.Frame(ctrlFrame, w=22).pack(side='left')
        # 复选框
        ctrlF1 = tk.Frame(ctrlFrame)
        ctrlF1.pack(side='left')
        self.isAutoOCR = tk.BooleanVar()
        self.isAutoOCR.set(True)
        wid = ttk.Checkbutton(
            ctrlF1, variable=self.isAutoOCR, text='启用 OCR结果预览')
        wid.grid(column=0, row=0, sticky='w')
        self.balloon.bind(wid, '以虚线框标出OCR识别到的文本块')
        wid = ttk.Checkbutton(
            ctrlF1, variable=Config.getTK('isAreaWinAutoTbpu'), text='启用 文块后处理预览')
        Config.addTrace('isAreaWinAutoTbpu', self.reLoadImage)
        wid.grid(column=0, row=1, sticky='w')
        self.balloon.bind(
            wid, '以虚线框标出经过文本块后处理的块\n注意，仅用于预览后处理效果，\n实际任务时忽略区域早于后处理执行，不受后处理的影响')
        Widget.comboboxFrame(ctrlF1, '', 'tbpu', width=18).grid(
            column=0, row=2, sticky='w')
        # 切换画笔按钮
        tk.Frame(ctrlFrame, w=30).pack(side='left')
        ctrlF2 = tk.Frame(ctrlFrame)
        ctrlF2.pack(side='left')
        self.buttons = [None, None, None]
        btnW = 15
        self.buttons[0] = tk.Button(ctrlF2, text='+忽略区域 A', width=btnW, height=3, fg=self.areaColor[0], bg=self.areaColor[3],
                                    command=lambda: self.changeMode(0))
        self.buttons[0].pack(side='left', padx=5)
        self.balloon.bind(
            self.buttons[0], '处于 [忽略区域 A] 内的矩形虚线文字块将被忽略\n一般情况下，将需要去除的水印区域全部画上区域A即可')
        tk.Frame(ctrlFrame, w=20).pack(side='left')
        self.buttons[1] = tk.Button(ctrlF2, text='+识别区域', width=btnW, height=3, fg=self.areaColor[1], bg=self.areaColor[3],
                                    command=lambda: self.changeMode(1))
        self.buttons[1].pack(side='left', padx=5)
        self.balloon.bind(
            self.buttons[1], '若 [识别区域] 内存在文字块，则 [忽略区域 A] 失效')
        tk.Frame(ctrlFrame, w=20).pack(side='left')
        self.buttons[2] = tk.Button(ctrlF2, text='+忽略区域 B', width=btnW, height=3, fg=self.areaColor[2], bg=self.areaColor[3],
                                    command=lambda: self.changeMode(2))
        self.buttons[2].pack(side='left', padx=5)
        self.balloon.bind(
            self.buttons[2], '当 [忽略区域 A] 失效，即触发 [识别区域] 时，\n[忽略区域 B] 生效')

        tk.Frame(ctrlFrame, w=20).pack(side='left')
        ctrlF4 = tk.Frame(ctrlFrame)
        ctrlF4.pack(side='left')
        tk.Button(ctrlF4, text='清空', width=12, bg='white',
                  command=self.clearCanvas).pack()
        tk.Button(ctrlF4, text='撤销\nCtrl+Shift+Z', width=12, bg='white',
                  command=self.revokeCanvas).pack(pady=5)
        self.win.bind("<Control-Z>", self.revokeCanvas)  # 绑定撤销组合键，带shift
        tk.Frame(ctrlFrame, w=10).pack(side='left')
        tk.Button(ctrlFrame, text='完成', width=8, bg='white',
                  command=lambda: self.onClose(False)).pack(side="left", fill="y")
        tipsFrame = tk.Frame(self.win)
        tipsFrame.pack()
        tk.Label(tipsFrame, text="↓ ↓ ↓ 将任意图片拖入下方预览。然后点击按钮切换画笔，在图片上按住左键框选出想要的区域。",
                 fg='gray').pack(side='left')
        tk.Label(tipsFrame, text="同一种区域可绘制多个方框。", fg='blue').pack(side='left')
        tk.Label(tipsFrame, text="↓ ↓ ↓", fg='gray').pack(side='left')
        # initPanel()

        # def initCanvas():  # 初始化绘图板
        self.canvas = tk.Canvas(self.win, width=self.cW, height=self.cH,
                                bg="gray", cursor="plus", borderwidth=0)
        self.canvas.bind("<Button-1>", self.mouseDown)  # 鼠标点击
        self.canvas.bind("<B1-Motion>", self.mouseMove)  # 鼠标移动
        self.canvas.pack()
        hook_dropfiles(self.win, func=self.draggedFiles)  # 注册文件拖入
        self.canvasImg = None  # 当前在显示的图片
        # initCanvas()

        def initOCR():  # 初始化识别器
            canvasText = self.canvas.create_text(self.cW/2, self.cH/2, font=('', 15, 'bold'), fill='white', anchor="c",
                                                 text=f'引擎启动中，请稍候……')
            try:
                OCRe.start()  # 启动或刷新引擎
            except Exception as e:
                tk.messagebox.showerror(
                    '遇到了亿点小问题',
                    f'识别器初始化失败：{e}\n\n请检查配置有无问题！')
                self.win.attributes('-topmost', 1)  # 设置层级最前
                self.win.attributes('-topmost', 0)  # 然后立刻解除
                self.isAutoOCR.set(False)  # 关闭自动分析
                return
            finally:
                try:
                    self.canvas.delete(canvasText)  # 删除提示文字
                except:
                    pass
        initOCR()

        if defaultPath:  # 打开默认图片
            self.loadImage(defaultPath)

    def onClose(self, isAsk=True):  # 点击关闭。isAsk为T时询问。

        def getData():  # 将数据传给接口，然后关闭窗口
            area = [[], [], []]
            for i in range(3):
                for a in self.area[i]:
                    a00, a01, a10, a11 = round(
                        a[0][0]/self.imgScale), round(a[0][1]/self.imgScale), round(a[1][0]/self.imgScale), round(a[1][1]/self.imgScale)
                    if a00 > a10:  # x对调
                        a00, a10 = a10, a00
                    if a01 > a11:  # y对调
                        a01, a11 = a11, a01
                    area[i].append([(a00, a01), (a10, a11)])
            return {"size": self.imgSize, "area": area}

        if self.area[0] or self.area[1] or self.area[2]:  # 数据存在
            if isAsk:  # 需要问
                if tk.messagebox.askokcancel('关闭窗口', '要应用选区吗？'):  # 需要应用
                    Config.set("ignoreArea", getData())
            else:  # 不需要问
                Config.set("ignoreArea", getData())
        if self.closeSendData:  # 通信接口存在，则回传数据
            self.closeSendData()
        OCRe.stopByMode()  # 关闭OCR进程
        self.win.destroy()  # 销毁窗口

    def draggedFiles(self, paths):  # 拖入文件
        self.loadImage(paths[0].decode(  # 根据系统编码来解码
            Config.sysEncoding, errors='ignore'))

    def reLoadImage(self):  # 载入旧图片
        if self.lastPath:
            self.loadImage(self.lastPath)

    def loadImage(self, path):  # 载入新图片
        """载入图片"""
        try:
            img = Image.open(path)
        except Exception as e:
            tk.messagebox.showwarning(
                "遇到了一点小问题", f"图片载入失败。图片地址：\n{path}\n\n错误信息：\n{e}")
            self.win.attributes('-topmost', 1)  # 设置层级最前
            self.win.attributes('-topmost', 0)  # 然后立刻解除
            return
        # 检测图片大小
        if self.imgSize == (-1, -1):  # 初次设定
            self.imgSize = img.size
            # 计算 按宽和高 分别缩放的比例
            sw, sh = self.cW/img.size[0], self.cH/img.size[1]
            # 测试，按宽、还是高缩放，刚好填满画布
            if sw > sh:  # 按高
                self.imgReSize = (round(img.size[0]*sh), self.cH)
                self.imgScale = sh
            else:  # 按宽
                self.imgReSize = (self.cW, round(img.size[1]*sw))
                self.imgScale = sw
            self.imgSizeText.set(f'{self.imgSize[0]}x{self.imgSize[1]}')
        elif not self.imgSize == img.size:  # 尺寸不符合
            tk.messagebox.showwarning("图片尺寸错误！",
                                      f"当前图像尺寸限制为{self.imgSize[0]}x{self.imgSize[1]}，不允许加载{img.size[0]}x{img.size[1]}的图片。\n若要解除限制、更改为其他分辨率，请点击“清空”后重新拖入图片。")
            self.win.attributes('-topmost', 1)  # 设置层级最前
            self.win.attributes('-topmost', 0)  # 然后立刻解除
            return
        self.clearCanvasImage()  # 清理上次绘制的图像

        # OCR识别
        def runOCR():
            # 任务前：显示提示信息
            self.win.title(f"分析中…………")  # 改变标题
            pathStr = path if len(
                path) <= 50 else path[:50]+"……"  # 路径太长显示不全，截取
            canvasText = self.canvas.create_text(self.cW/2, self.cH/2, font=('', 15, 'bold'), fill='white', anchor="c",
                                                 text=f'图片分析中，请稍候……\n\n\n\n{pathStr}')
            self.win.update()  # 刷新窗口
            # 开始识别，耗时长
            data = OCRe.run(path)
            # 任务后：刷新提示信息
            self.canvas.delete(canvasText)  # 删除提示文字
            if data['code'] == 100:  # 存在内容
                if Config.get('isAreaWinAutoTbpu'):  # 需要后处理
                    tbpuClass = Config.get('tbpu').get(
                        Config.get('tbpuName'), None)
                    if tbpuClass:
                        name = os.path.basename(path)  # 带后缀的文件名
                        imgInfo = {'name': name,
                                   'path': path, 'size': img.size}
                        tbpu = tbpuClass()
                        data['data'], s = tbpu.run(data['data'], imgInfo)
                for o in data["data"]:  # 绘制矩形框
                    # 提取左上角、右下角的坐标
                    p1x = round(o['box'][0][0]*self.imgScale)+self.tran
                    p1y = round(o['box'][0][1]*self.imgScale)+self.tran
                    p2x = round(o['box'][2][0]*self.imgScale)+self.tran
                    p2y = round(o['box'][2][1]*self.imgScale)+self.tran
                    r1 = self.canvas.create_rectangle(
                        p1x, p1y, p2x, p2y, outline='white', width=2)  # 绘制白实线基底
                    r2 = self.canvas.create_rectangle(
                        p1x, p1y, p2x, p2y, outline='black', width=2, dash=4)  # 绘制黑虚线表层
                    self.canvas.tag_lower(r2)  # 移动到最下方
                    self.canvas.tag_lower(r1)
                    self.areaTextRec.append(r1)
                    self.areaTextRec.append(r2)
            elif not data["code"] == 101:  # 发生异常
                self.isAutoOCR.set(False)  # 关闭自动分析
                tk.messagebox.showwarning(
                    "遇到了一点小问题", f"图片分析失败。图片地址：\n{path}\n\n错误码：{str(data['code'])}\n\n错误信息：\n{str(data['data'])}")
                self.win.attributes('-topmost', 1)  # 设置层级最前
                self.win.attributes('-topmost', 0)  # 然后立刻解除
        if self.isAutoOCR.get():
            try:
                runOCR()
            except Exception as e:
                tk.messagebox.showerror('遇到了一点小问题', f'预览OCR失败：\n{e}')
                return
        self.win.title(f"选择区域 {path}")  # 改变标题
        # 缓存图片并显示
        img = img.resize(self.imgReSize, Image.ANTIALIAS)  # 改变图片大小
        self.imgFile = ImageTk.PhotoImage(img)  # 缓存图片
        self.canvasImg = self.canvas.create_image(
            0, 0, anchor='nw', image=self.imgFile)  # 绘制图片
        self.canvas.tag_lower(self.canvasImg)  # 该元素移动到最下方，防止挡住矩形们
        self.lastPath = path

    def changeMode(self, type_):  # 切换绘图模式
        if self.imgSize == (-1, -1):
            return
        self.areaType = type_
        for b in self.buttons:
            b["state"] = tk.NORMAL  # 启用所有按钮
            b["bg"] = self.areaColor[3]
        self.buttons[type_]["state"] = tk.DISABLED  # 禁用刚按下的按钮
        self.buttons[type_]["bg"] = self.areaColor[type_]  # 切换背景颜色

    def clearCanvasImage(self):  # 清除画布上的图片其文字框，不清理忽略框
        self.lastPath = ''
        if self.canvasImg:  # 删除图片
            self.canvas.delete(self.canvasImg)
        for t in self.areaTextRec:  # 删除文字框
            self.canvas.delete(t)
        self.areaTextRec = []

    def clearCanvas(self):  # 清除画布
        self.lastPath = ''
        self.win.title(f"选择区域")  # 改变标题
        self.area = [[], [], []]
        self.areaHistory = []
        self.areaTextRec = []
        self.areaType = -1
        self.areaTypeIndex = [-1, -1, -1]
        self.imgSize = (-1, -1)
        self.imgSizeText.set("未设定")
        self.canvas.delete(tk.ALL)
        for b in self.buttons:
            b["state"] = tk.NORMAL  # 启用所有按钮
            b["bg"] = self.areaColor[3]

    def revokeCanvas(self, e=None):  # 撤销一步
        if len(self.areaHistory) == 0:
            return
        self.canvas.delete(self.areaHistory[-1]["id"])  # 删除历史记录中最后绘制的矩形
        self.area[self.areaHistory[-1]["type"]].pop()  # 删除对应类型的最后一个矩形数据
        self.areaHistory.pop()  # 删除历史记录

    def mouseDown(self, event):  # 鼠标按下，生成新矩形
        if self.areaType == -1:
            return
        x, y = event.x, event.y
        if x > self.imgReSize[0]:  # 防止越界
            x = self.imgReSize[0]
        elif x < 0:
            x = 0
        if y > self.imgReSize[1]:
            y = self.imgReSize[1]
        elif y < 0:
            y = 0
        c = self.areaColor[self.areaType]
        id = self.canvas.create_rectangle(  # 绘制新图
            x+self.tran, y+self.tran, x+self.tran, y+self.tran,  width=2, activefill=c, outline=c)
        self.area[self.areaType].append([(x, y), (x, y)])  # 向对应列表中增加1个矩形
        self.areaHistory.append({"id": id, "type": self.areaType})  # 载入历史记录

    def mouseMove(self, event):  # 鼠标拖动，刷新最后的矩形
        if self.areaType == -1:
            return
        x, y = event.x, event.y
        if x > self.imgReSize[0]:  # 防止越界
            x = self.imgReSize[0]
        elif x < 0:
            x = 0
        if y > self.imgReSize[1]:
            y = self.imgReSize[1]
        elif y < 0:
            y = 0
        x0, y0 = self.area[self.areaType][-1][0][0], self.area[self.areaType][-1][0][1]
        # 刷新历史记录中最后的图形的坐标
        self.canvas.coords(
            self.areaHistory[-1]["id"], x0+self.tran, y0+self.tran, x+self.tran, y+self.tran)
        self.area[self.areaType][-1][1] = (x, y)  # 刷新右下角点



from utils.config import Config

import tkinter as tk
from enum import Enum

winW, winH = 300, 60  # 宽高
winR = 15  # 圆角半径
winPosY = 60  # 窗口与屏幕下方的距离
xxS1 = 14  # 叉叉进度条大小（直径）
xxS2 = 10  # 叉叉进度条离右上角的距离
bgColor = '#FFCBE0'  # 背景色
teS1 = 12  # 主标题字号
teS2 = 9  # 次内容字号
teP1 = (12, 9)  # 主标题位置
teP2 = (12, 35)  # 次内容位置
popTime = 0.4  # 弹出动画时长，秒
downTime = 6  # 自动关闭倒计时时长，秒
frameTime = 30  # 动画帧间隔，毫秒
winR2 = winR*2


class State(Enum):
    none = 0  # 无
    starting = 1  # 弹出中
    showing = 2  # 显示中


class NotifyWindow():
    def __init__(self):
        self.win = None
        self.state = State.none
        self.afters = [None, None, None]  # 保存运行中的计时器

    def __initWin(self):  # 初始化窗体
        self.win = tk.Toplevel()
        self.win.withdraw()
        self.win.attributes('-alpha', 0)  # 透明度

        self.__winGO()  # 初始化窗口位置
        self.win.resizable(False, False)  # 锁定大小
        self.win.overrideredirect(True)  # 无边框模式
        self.win['bg'] = '#FFFFFF'
        self.win.attributes('-transparentcolor', '#FFFFFF')  # 设置纯白色为透明
        # 画布
        self.canvas = tk.Canvas(self.win, bg='#FFFFFF', highlightthickness=0)
        self.canvas.pack(fill=tk.BOTH, expand=True)
        self.canvas.bind('<Button-1>', self.__onClick)
        # 绘制圆角矩形
        self.canvas.create_oval(
            0, 0, winR2, winR2, fill=bgColor, outline='')
        self.canvas.create_oval(
            winW-winR2, 0, winW, winR2, fill=bgColor, outline='')
        self.canvas.create_oval(
            winW-winR2, winH-winR2, winW, winH, fill=bgColor, outline='')
        self.canvas.create_oval(
            0, winH-winR2, winR2, winH, fill=bgColor, outline='')
        self.canvas.create_rectangle(
            winR, 0, winW-winR, winH, fill=bgColor, outline='')
        self.canvas.create_rectangle(
            0, winR, winW, winH-winR, fill=bgColor, outline='')
        # 绘制叉叉进度条
        self.prog = self.canvas.create_arc(winW-xxS1-xxS2, xxS2, winW-xxS2, xxS1+xxS2,
                                           fill='', outline='gray', width=1, style='arc',
                                           start=90, extent=0)
        # self.canvas.create_text(
        #     winW-xxS2-int(xxS1/2), xxS2+int(xxS1/2), anchor='center',
        #     font=('Microsoft YaHei', xxS1-xxS2), text='×')
        # 绘制文本
        self.text1 = self.canvas.create_text(
            *teP1, anchor='nw', font=('Microsoft YaHei', teS1),
            text='')
        self.text2 = self.canvas.create_text(
            *teP2, anchor='nw', font=('Microsoft YaHei', teS2), fill='#666',
            text='')

    @staticmethod
    def __easing(i):  # 0-1之间的缓动函数，越靠近0增长越快
        return 1+(i-1)*(i-1)*(i-1)

    def __winGO(self, yValue=0):  # 移动窗口
        winX = int((self.win.winfo_screenwidth() - winW) / 2)  # 位置：屏幕中下
        winY = int(self.win.winfo_screenheight() - winH - yValue)
        self.win.geometry(f'{winW}x{winH}+{winX}+{winY}')

    def __actionStart(self, t=popTime):  # 开始弹alpha出动画
        if t <= 0:
            self.state = State.showing  # 状态：显示中
            return
        p = (popTime-t)/popTime  # 0→1
        p = self.__easing(p)  # 缓动
        y = winPosY * p
        self.__winGO(y)  # 移动窗口
        self.win.attributes('-alpha', p)  # 透明度
        self.afters[0] = self.win.after(
            frameTime, lambda: self.__actionStart(t-frameTime/1000))

    def __actionEnd(self, t=popTime):  # 结束弹回动画
        if t <= 0:
            self.close()
            return
        p = t/popTime  # 1→0
        p = self.__easing(p)  # 缓动
        y = winPosY * p
        self.__winGO(y)  # 移动窗口
        self.win.attributes('-alpha', p)  # 透明度
        self.afters[1] = self.win.after(
            frameTime, lambda: self.__actionEnd(t-frameTime/1000))

    def __actionCountdown(self, t=downTime):  # 倒计时动画
        if t <= 0:
            self.__actionEnd()
            return
        extent = 360 * (t/downTime)  # 360→0
        if extent < 5:
            extent = 0
        self.canvas.itemconfig(self.prog, extent=extent)
        self.afters[2] = self.win.after(
            frameTime, lambda: self.__actionCountdown(t-frameTime/1000))

    def __onClick(self, *e):  # 点击事件
        self.__afterCancel()  # 取消所有运行中的计时器
        self.canvas.itemconfig(self.prog, extent=0)  # 倒计时复位
        self.__actionEnd()  # 动画关闭

    def __afterCancel(self):  # 取消所有运行中的计时器
        for i, a in enumerate(self.afters):
            if a:
                self.win.after_cancel(a)
                self.afters[i] = None

    def show(self, title, message):
        '''显示一条消息。'''
        if not self.win:
            self.__initWin()
        if not self.state == State.none:
            self.close()  # 关闭上一条消息
        self.state = State.starting  # 状态：弹出中
        # 初始化参数
        self.__winGO()  # 位置
        self.win.attributes('-alpha', 0)  # 透明度
        self.canvas.itemconfig(self.prog, extent=0)  # 进度条
        # 修改文字
        title = title.replace('\n', '')
        message = message.replace('\n', '')
        self.canvas.itemconfig(self.text1, text=title)
        self.canvas.itemconfig(self.text2, text=message)
        # 恢复窗口
        self.win.attributes('-topmost', 1)  # 设置层级最前
        self.win.state('normal')  # 恢复前台状态
        # 弹出动画
        self.__actionStart()
        # 启动倒计时动画
        if downTime > 0:
            self.__actionCountdown()

    def close(self):
        '''关闭当前通知'''
        if self.state == State.none:
            return
        self.win.withdraw()  # 隐藏消息
        self.__afterCancel()  # 取消所有运行中的计时器
        self.state = State.none


NotifyWin = NotifyWindow()


def Notify(title, msg):
    if not Config.get('isNotify'):
        return
    NotifyWin.show(title, msg)


def NotifyClose():
    NotifyWin.close()



# 封装 pynput ，提供对外 Hotkey 接口
from pynput import keyboard, mouse
from pynput._util.win32 import KeyTranslator

from time import time

from utils.config import Config


class KeyTranslator_Api:  # 封装 keyTranslator ，负责key、char、vk的转换

    def __init__(self):
        self._kt = KeyTranslator()
        self._layout, _layoutData = self._kt._generate_layout()
        self._normalLayout = _layoutData[(  # 选取常规布局，不受修饰键影响
            False, False, False)]

    def __call__(self, key):
        '''传入pynput的Key对象，返回与修饰键无关的键名char'''
        # 比如，就算按下Shift再按“=”，依然返回“=”而不是“+”
        try:
            if hasattr(key, 'name'):  # 若为控制键
                name = key.name.replace('cmd', 'win')  # win键名称修正
                if '_' in name:  # 清除 _l _r 标记后缀
                    return name.split('_', 1)[0].lower()
                return name.lower()
            else:  # 若为非控制键，通过vk获取键名，避免组合键的char为空
                scan = self._kt._to_scan(key.vk, self._layout)  # vk转扫描码
                char = self._normalLayout[scan][0]  # 扫描码转char
                return char.lower()
        except Exception as e:  # 特殊键（如Fn）没有对应字符，会跳到这里
            if key and hasattr(key, 'vk'):
                return f'<{key.vk}>'  # 未知键值，无对应字符，返回键值本身
            else:
                return ''

    def char2vk(self, char):
        '''传入控制键字符串char，返回对应的vk值'''
        if hasattr(keyboard.Key, char):
            vk = getattr(keyboard.Key, char).value
            return vk
        else:  # 非控制键返回自己
            return char


KTA = KeyTranslator_Api()


class Hotkey_Api():  # 热键API，封装 keyboard.Listener

    # 记录一个键按下状态的类
    class Press_Key():

        def __init__(self, keyName):
            self.keyName = keyName
            # TTL：生存时间，秒。一个键按下超过此时间后将忽略它，直到它下次被按下。
            # 这是为了防止意外没有接收到一个键的释放事件，导致它长期留在已按字典里，引起组合键误触。
            self.TTL = 10
            self.updateTTL()

        def updateTTL(self):
            '''刷新生存时间'''
            self.TTL = time()+Config.get('hotkeyMaxTtl')

        def isLive(self, nowTime=None):
            '''若此键仍存活，返回T'''
            if nowTime:
                return nowTime < self.TTL
            return time() < self.TTL

    # 记录一组热键状态的类
    class Hot_Key():

        def __init__(self, hotkey, callback, isPress=True):
            # 通过字符串创建
            if isinstance(hotkey, str):
                self.hotkeyName = hotkey  # 名称
                self.hotkeySet = self.toSet(hotkey)  # 键集合
            # 通过集合创建
            elif isinstance(hotkey, set):
                self.hotkeySet = hotkey  # 键集合
                self.hotkeyName = self.toStr(hotkey)  # 名称
            else:
                raise ValueError(f'不合法的热键值：{hotkey}')
            self.callback = callback  # 回调
            self.isPress = isPress  # T为按压触发，F为释放触发

        @staticmethod
        def toSet(hotkey):
            '''传入热键列表或集合，返回组合键字符串'''
            return set(hotkey.split('+'))

        @staticmethod
        def toStr(hotkeyName):
            '''传入热键名字符串，返回键集'''
            return '+'.join(hotkeyName)

        def isNameEQ(self, hotkeyName):
            '''传入热键字符串，判断是否与本键相等'''
            return self.hotkeySet == self.toSet(hotkeyName)

        def isSetEQ(self, hotkeySet):
            '''传入热键集合，判断是否与本键相等'''
            return self.hotkeySet == hotkeySet

        def isSetSub(self, hotkeySet):
            '''传入热键集合，判断是否包含本键'''
            return self.hotkeySet.issubset(hotkeySet)

        def isKeyIn(self, keyName):
            '''传入单个按键，判断是否在本热键组合内'''
            return keyName in self.hotkeySet

    # ==============================================================

    def __init__(self):
        self.pressDict = {}  # 已按字典
        self.hotkeyList = []  # 热键列表
        # 监听键盘
        self.listener = keyboard.Listener(
            on_press=self._onPress,
            on_release=self._onRelease)
        self.listener.start()
        self.controller = keyboard.Controller()
        self.isReading = False  # 录制模式
        self.readData = {  # 录制信息
            'keyList': [],  # 本组录制的结果
            'keySet': set(),  # 当前等待释放的录制键
            'callback': None  # 录制完成的回调
        }

        # 监听鼠标
        self.mouseListener = None
        self.isMouseListener = False  # 记录鼠标监听器是否在运行
        self.mouseCallback = {'up': None, 'down': None}  # 鼠标回调
        # self.mouseListener.start()
        self.mouseController = mouse.Controller()  # 鼠标操控器

    def join(self):
        '''调试用，键盘监听器运行期间阻塞线程'''
        self.listener.join()

    def _onPress(self, key_):  # 一个键被按下的回调
        self._checkTTL()  # 检查TTL，移除长久没有释放的异常键
        keyName = KTA(key_)  # 键名字符串
        # 维护已按字典
        if keyName not in self.pressDict:  # 按下按键，则加入已按字典
            self.pressDict[keyName] = self.Press_Key(keyName)
        else:   # 此键重复按下（未监听到上次释放），则刷新生存时间
            self.pressDict[keyName].updateTTL()
        # 录制
        if self.isReading:
            if not keyName in self.readData['keyList']:  # 不允许重复录
                self.readData['keyList'].append(keyName)
                self.readData['keySet'].add(keyName)
        # 触发热键
        else:
            self._callHotkey(keyName, True)

    def _onRelease(self, key_):  # 一个键被释放的回调
        self._checkTTL()  # 检查TTL，移除长久没有释放的异常键
        keyName = KTA(key_)  # 键名字符串
        # 录制
        if self.isReading:
            self.readData['keySet'].discard(keyName)  # 删除
            if not self.readData['keySet']:  # 集合为空，所有按下的按键都释放，结束录制
                keyStr = self.Hot_Key.toStr(self.readData['keyList'])
                self.isReading = False
                self.readData['callback'](keyStr)
        # 触发热键
        else:
            self._callHotkey(keyName, False)
        # 维护已按字典
        if keyName in self.pressDict:  # 在已按字典中
            del self.pressDict[keyName]  # 移出已按字典

    def _checkTTL(self):  # 检测已按字典的存活性
        nowTime = time()
        for k, v in list(self.pressDict.items()):
            if not v.isLive(nowTime):
                del self.pressDict[k]

    def _callHotkey(self, key, isPress):  # 检测并触发热键
        nowKeySet = set(self.pressDict.keys())
        debugMsg = f'🖮 {"↓ " if isPress else " ↑"} {key} | {",".join(nowKeySet)}'
        flag = Config.get('isHotkeyStrict')
        for hk in self.hotkeyList:
            if hk.isKeyIn(key) and hk.isPress == isPress:
                # 严格模式，已按集合必须与热键集合完全一致才能触发
                if flag and hk.isSetEQ(nowKeySet):
                    hk.callback()
                    debugMsg += f' | √ {hk.hotkeyName}'
                # 宽容模式，已按集合包含热键集合即可触发
                elif not flag and hk.isSetSub(nowKeySet):
                    hk.callback()
                    debugMsg += f' | √ {hk.hotkeyName}'
        if Config.get('isDebug'):
            print(debugMsg)

    def _addMouseListener(self):

        def _onMouseClick(x, y, button, pressed):  # 鼠标按钮回调
            print(x, y, button, pressed)
            if self.mouseCallback['down'] and pressed:
                self.mouseCallback['down']((x, y))
            elif self.mouseCallback['up'] and not pressed:
                self.mouseCallback['up']((x, y))

        if not self.isMouseListener:
            self.mouseListener = mouse.Listener(on_click=_onMouseClick)
            self.mouseListener.start()
            self.isMouseListener = True

    # ======================= 键盘对外接口 =============================

    def add(self, hotkey, callback, isPress=True):
        '''添加一个快捷键组合监听。按下时调用callback'''
        hotkey = hotkey.lower()  # 转小写
        # 判断重复
        for h in self.hotkeyList:
            if h.isNameEQ(hotkey) and h.callback.__name__ == callback.__name__ \
                    and h.isPress == isPress:
                return
        hk = self.Hot_Key(hotkey, callback, isPress)
        self.hotkeyList.append(hk)

    def addRelease(self, hotkey, callback):
        '''添加一个快捷键组合监听。释放时调用callback'''
        self.add(hotkey, callback, False)

    def remove(self, hotkey):
        '''移除一组快捷键对应的所有监听事件'''
        hset = self.Hot_Key.toSet(hotkey)
        for i in range(len(self.hotkeyList)-1, -1, -1):
            hk = self.hotkeyList[i]
            if hk.isSetEQ(hset):
                del self.hotkeyList[i]

    def isPressed(self, hotkey):
        '''检查当前按键序列是否包含hotkey，是返回T'''
        hset = self.Hot_Key.toSet(hotkey)
        nowKeySet = set(self.pressDict.keys())
        return hset.issubset(nowKeySet)

    def send(self, hotkey):
        '''发送按键序列'''
        keynameList = hotkey.split('+')
        vkList = []
        for name in keynameList:
            name = name.replace('win', 'cmd')
            vk = KTA.char2vk(name)
            vkList.append(vk)
        vks = tuple(vkList)
        with self.controller.pressed(*vks):  # 按需按下和释放
            pass
        # 手动在已按字典中移除
        for name in keynameList:
            if name in self.pressDict:
                del self.pressDict[name]

    def read(self, callback):
        '''录制快捷键。按下并松开一组按键，将按键序列字符串发送到回调函数'''
        if self.isReading:
            callback('', '当前已在录制')
            return
        self.isReading = True
        self.readData['keyList'] = []
        self.readData['keySet'] = set()
        self.readData['callback'] = callback

    # ======================= 鼠标对外接口 =============================

    def addMouseButtonDown(self, callback):
        '''添加一个鼠标按钮监听。按下时调用callback，返回xy坐标元组'''
        self._addMouseListener()
        self.mouseCallback['down'] = callback

    def addMouseButtonUp(self, callback):
        '''添加一个鼠标按钮监听。松开时调用callback，返回xy坐标元组'''
        self._addMouseListener()
        self.mouseCallback['up'] = callback

    def removeMouse(self):
        '''移除鼠标监听'''
        if self.isMouseListener:
            self.mouseListener.stop()
            self.isMouseListener = False
            self.mouseCallback = {'up': None, 'down': None}

    def getMousePos(self):
        return self.mouseController.position

    def setMousePos(self, pos):
        '''设置鼠标位置'''
        self.mouseController.position = pos


hotkeyApi = Hotkey_Api()



from utils.pynput_hotkey import hotkeyApi
from utils.logger import GetLog


Log = GetLog()


class Hotkey():  # 热键

    # ========================== 键盘 ==============================

    @staticmethod
    def add(hotkey, callback):
        '''添加一个快捷键组合监听。按下时调用callback'''
        # suppress=False 允许按键事件继续向别的软件下发，否则系统中相同辅助键的快捷键全都失效
        hotkeyApi.add(hotkey, callback)
        Log.info(f'注册按下 ↓【{hotkey}】')

    @staticmethod
    def addRelease(hotkey, callback):
        '''添加一个快捷键组合监听。松开时调用callback'''
        hotkeyApi.addRelease(hotkey, callback)
        Log.info(f'注册抬起 ↑【{hotkey}】')

    @staticmethod
    def remove(hotkey):
        '''移除一个快捷键组合监听'''
        hotkeyApi.remove(hotkey)  # 移除该快捷键
        # Log.info(f'注销监听 ×【{hotkey}】')

    @staticmethod
    def read(callback):
        '''录制快捷键。按下并松开一组按键，将按键序列字符串发送到回调函数\n
            成功：返回hotkey字符串\n
            失败：hotkey为空，errmsg为错误字符串'''
        hotkeyApi.read(callback)

    @staticmethod
    def isPressed(hotkey):
        '''检查当前是否按下hotkey序列，是返回T'''
        return hotkeyApi.isPressed(hotkey)

    @staticmethod
    def send(hotkey):
        '''发送按键序列'''
        hotkeyApi.send(hotkey)
        Log.info(f'发送按键组合【{hotkey}】')

    # ========================== 鼠标 ==============================

    @staticmethod
    def addMouseButtonDown(callback):
        '''添加一个鼠标按钮监听。按下时调用callback，返回xy坐标元组'''
        hotkeyApi.addMouseButtonDown(callback)

    @staticmethod
    def addMouseButtonUp(callback):
        '''添加一个鼠标按钮监听。松开时调用callback，返回xy坐标元组'''
        hotkeyApi.addMouseButtonUp(callback)

    @staticmethod
    def removeMouse():
        '''移除鼠标监听'''
        hotkeyApi.removeMouse()

    @staticmethod
    def getMousePos():
        '''获取鼠标当前位置'''
        return hotkeyApi.getMousePos()

    @staticmethod
    def setMousePos(pos):
        '''设置鼠标当前位置'''
        return hotkeyApi.setMousePos(pos)



# 文块处理：横排-单行-模糊匹配
from ocr.tbpu.merge_line_h import TbpuLineH


class TbpuLineHFuzzy(TbpuLineH):
    def __init__(self):
        super().__init__()
        self.tbpuName = '横排-单行-模糊匹配'
        self.isLimitX = False  # 不考虑文块水平距离

    def merge2text(self, text1, text2):
        '''合并两段文字的规则'''
        return text1 + ' ' + text2



from utils.logger import GetLog
from utils.config import Config, ScsModeFlag
from utils.tool import Tool
from utils.hotkey import Hotkey  # 快捷键
from ui.win_notify import NotifyClose  # 关闭通知弹窗
from ui.win_show_image import ShowImage  # 显示图片窗口

# 获取显示器信息
from win32api import EnumDisplayMonitors, GetMonitorInfo
from win32gui import CreateDC
from win32print import GetDeviceCaps
# 剪贴板
from io import BytesIO
from win32clipboard import OpenClipboard, EmptyClipboard, SetClipboardData, CloseClipboard, CF_DIB

import tkinter as tk
from PIL import ImageGrab, ImageTk
from enum import Enum

# TODO :
# 截图模块的工作原理是：先获取虚拟屏幕（所有显示器的画面拼凑在一起）的完整截图。然后创建一块画布，
# 画布的起点(左上角)是虚拟屏幕的左上角(xy可能为负值)，画布的宽高是虚拟屏幕的宽高，
# 然后在画布上显示完整截图，监听用户按下与拖拽。此时画布上显示的截图的位置应与真实屏幕画面一一对应。
# 但问题是，画布作为一个窗口，自身有一个缩放比例（即它出生的那块屏幕的缩放比）。
# 若多屏幕的其中某块屏幕的缩放比与画布的缩放比不一致，在一定排列下，它们的逻辑坐标会错位，表现为画面错位。
# 我原本希望通过获取全部屏幕的物理分辨率和逻辑分辨率，得到各自的缩放比，进而计算出画布坐标系下的“真实”逻辑坐标。
# 但是，我没能摸清规律。参数与太多因素相关，
# 屏幕的排列方式、软件点开时的状态、画布创建时的位置……都可能影响逻辑坐标和逻辑分辨率，
# 使得很难计算画布坐标系下应该对屏幕矫正的锚点和比例。

Log = GetLog()


def _ScreenshotClose(flag, errMsg=None):
    Log.info('截图结束')
    Config.main.closeScreenshot(flag, errMsg)


def ScreenshotCopy():
    '''截屏，保存到剪贴板，然后调用主窗的closeScreenshot接口'''
    NotifyClose()  # 关闭通知弹窗
    scsMode = Config.get('scsMode').get(Config.get(
        'scsModeName'), ScsModeFlag.multi)  # 当前截屏模式
    if scsMode == ScsModeFlag.multi:  # 内置截图模式
        SSWin.startGrab()
    elif scsMode == ScsModeFlag.system:  # 系统截图模式
        SSSys.startGrab()
    else:
        _ScreenshotClose(False, '未知的截图模式！')


class ScreenshotSys():  # 系统截图模式

    def __init__(self):
        self.isInitKey = False
        self.isWorking = False
        self.checkTimeMax = 10  # 最大检查次数
        self.checkTimeRate = 20  # 检查间隔频率，毫秒
        self.checkTime = 0  # 当前剩余检查次数
        self.position = (0, 0)

    def startGrab(self):  # 启动截屏
        '''启动系统截图。若通过快捷键进入，必须为win+shift+S'''
        Tool.emptyClipboard()  # 清空剪贴板
        self.isWorking = True
        if not self.isInitKey:
            self.__initKey()
        if not Hotkey.isPressed('win'):  # 不是通过快捷键进入
            Hotkey.send('win+shift+s')  # 发送系统截图快捷键
        Log.info('系统截图启动')

    def __initKey(self):  # 初始化监听
        # 绑定全局事件
        Hotkey.addRelease(  # Esc抬起，系统截图失败
            'esc', lambda: self.__close(False))
        Hotkey.addMouseButtonDown(self.__onDown)  # 注册监听鼠标左/右按下
        Hotkey.addMouseButtonUp(self.__onUp)  # 注册监听鼠标左/右抬起
        self.isInitKey = True

    def __onDown(self, pos):  # 用户操作开始
        if self.isWorking:
            self.position = pos  # 获取鼠标当前位置

    def __onUp(self, pos):  # 用户操作结束
        if self.isWorking:
            if self.position == pos:  # 鼠标起始结束位置相同，截图失败
                self.__close(False)
                return
            self.checkTime = 0
            self.__checkClipboard()

    def __checkClipboard(self):  # 检查剪贴板中是否已存在截图
        if self.checkTime >= self.checkTimeMax:
            self.__close(False, '读取剪贴板失败')  # 检查次数超限，截图失败
            return
        clipData = Tool.getClipboardFormat()  # 读取剪贴板
        if clipData == 2:  # 系统截图已保存到剪贴板内存，截图成功
            Log.info(f'  第{self.checkTime}次检查成功')
            if Config.get('isShowImage'):  # 显示图片展示窗
                ShowImage(imgPIL=ImageGrab.grabclipboard())
                self.__close(False)
            else:
                self.__close(True)
            return
        Log.info(f'  第{self.checkTime}次检查')
        self.checkTime += 1
        # 定时器指定下一轮查询
        Config.main.win.after(self.checkTimeRate, self.__checkClipboard)

    def __close(self, flag=False, errMsg=None):  # 退出
        if self.isWorking:
            Hotkey.removeMouse()  # 注销监听鼠标
            self.isInitKey = False
            self.isWorking = False
            _ScreenshotClose(flag, errMsg)


SSSys = ScreenshotSys()


class _DrawMode(Enum):
    ready = 1  # 准备中
    drag = 2  # 拖拽中


class ScreenshotWin():  # 内置截图模式
    OB = -100  # 元素隐藏屏幕外的位置

    def __init__(self):
        self.isInitWin = False  # 防止重复初始化窗体
        self.isInitGrab = False  # 防止未初始化截图参数时触发事件
        self.errMsg = None  # 记录错误，传给调用者
        self.screenScaleList = None  # 记录各个屏幕分别的缩放比例
        self.promptSss = True  # 本次使用期间显示缩放提示
        self.lastScInfos = None  # 上一轮的屏幕参数

    def startGrab(self):  # 启动截屏
        '''启动区域截图'''
        # “虚拟屏幕”指多显示器画面的拼凑在一起的完整画面
        self.image = ImageGrab.grab(all_screens=True)  # 对整个虚拟屏幕截图，物理分辨率

        if not self.isInitWin:
            self.__initWin()

        self.imageResult = None  # 结果图片
        self.sourceBox = None  # 截图包围盒原始信息
        self.drawMode = _DrawMode.ready  # 准备模式
        # 获取所有屏幕的信息，提取其中的坐标信息(虚拟，非物理分辨率)
        scInfos = EnumDisplayMonitors()  # 所有屏幕的信息
        self.scBoxList = [s[2] for s in scInfos]  # 提取虚拟分辨率的信息
        # 计算缩放比例，若不一致，则发送提示弹窗
        # 条件：需要提示 | 大于一块屏幕时 | 本次信息与上次不同 | 设置需要提示
        scInfosLen = len(scInfos)
        if self.promptSss and scInfosLen > 1 and not self.lastScInfos == scInfos and Config.get('promptScreenshotScale'):
            scList = []
            self.lastScInfos = scInfos  # 屏幕信息与上次一样时跳过检测，减少耗时
            # 提取所有屏幕缩放比例
            for index, sc in enumerate(scInfos):
                # 获取设备信息字典，得到设备名称 Device
                # 物理设备信息(dict) = GetMonitorInfo(hMonitor)
                info = GetMonitorInfo(scInfos[index][0])
                # 为显示设备创建设备上下文，得到物理设备句柄 hDC
                # 设备句柄(int) = CreateDC (设备名称, 设备名称 , None )
                hDC = CreateDC(info['Device'], info['Device'], None)
                w = GetDeviceCaps(hDC, 118)  # 常量 win32con.DESKTOPHORZRES
                # h = GetDeviceCaps(hDC, 117)  # 常量 win32con.DESKTOPVERTRES
                # 得到缩放比，即windows的“更改文本、应用等项目的大小”
                s = w / (sc[2][2]-sc[2][0])
                scList.append(s)
            # 检查缩放比例是否一致
            isEQ = True
            for i in range(1, scInfosLen):
                if not abs(scList[i] - scList[0]) < 0.001:
                    isEQ = False
                    break
            # 不一致，提示
            if not isEQ:
                self.screenScaleList = scList
                msg = f'''您当前使用{scInfosLen}块屏幕，且缩放比例不一致，分别为 {scList} 。

可能导致Umi-OCR截图异常，如画面不完整、窗口变形、识别不出文字等。
若出现这种情况，
请在系统设置【更改文本、应用等项目的大小】将所有屏幕调到相同数值。
或者，请在软件设置里将截图模式切换到【Windows 系统截图】。\n'''
                Config.main.panelOutput(msg)
                Config.main.notebook.select(
                    Config.main.notebookTab[1])  # 转到输出卡
                if tk.messagebox.askyesno('提示',
                                          f'{msg}\n本次使用不再提示此消息请点击[是]，永久不再提示请点击[否]'):
                    self.promptSss = False
                else:
                    Config.set('promptScreenshotScale', False, isSave=True)
        # 计算虚拟屏幕最左上角和最右下角的坐标
        scUp, scDown, scLeft, scRight = 0, 0, 0, 0
        for s in self.scBoxList:  # 遍历所有屏幕，获取最值
            if s[0] < scLeft:  # 左边缘
                scLeft = s[0]
            if s[1] < scUp:  # 上边缘
                scUp = s[1]
            if s[2] > scRight:  # 右边缘
                scRight = s[2]
            if s[3] > scDown:  # 下边缘
                scDown = s[3]
            # 计算虚拟屏幕的宽和高，请确保屏幕对齐
            scWidth, scHeight = scRight - scLeft, scDown - scUp
        # 多显示器处理完毕
        self.scBoxVirtual = (scLeft, scUp, scRight, scDown,
                             scWidth, scHeight)
        self.allScale = self.image.size[0] / scWidth  # 整个虚拟屏幕的缩放比例
        # 主窗口设置为铺满虚拟屏幕
        bd, bdp = 2, 1  # 边缘要额外拓展1像素，以免无法接收到鼠标在边缘的点击
        scStr = f'{scWidth+bd}x{scHeight+bd}+{scLeft-bdp}+{scUp-bdp}'
        # print(f'缩放比：{self.allScale}')
        # self.topwin.tk.call('tk', 'scaling', self.allScale/75)
        self.topwin.geometry(scStr)
        self.canvas['width'] = scWidth+bd
        self.canvas['height'] = scHeight+bd
        # 原图改物理为虚拟屏幕分辨率，转成tk格式，导入画布
        self.imageTK = ImageTk.PhotoImage(
            self.image.resize((scWidth, scHeight)))
        cimg = self.canvas.create_image(  # 底图
            bdp, bdp, anchor='nw', image=self.imageTK)
        self.canvas.lower(cimg)  # 移动到最下方

        self.topwin.deiconify()  # 显示窗口
        self.isInitGrab = True
        Log.info('初始化截图')
        self.__flash()  # 闪光
        if Config.get('isDebug'):  # 显示debug信息
            c = 2 if self.debugList else 1  # 若上一轮已显示，则调用两次以刷新
            for i in range(c):  # 否则，调用一次以打开
                self.__switchDebug()

    def __initWin(self):  # 初始化窗体
        self.isInitWin = True
        # 创建窗口
        self.topwin = tk.Toplevel()
        self.topwin.withdraw()  # 隐藏窗口
        self.topwin.overrideredirect(True)  # 无边框
        self.topwin.configure(bg='black')
        # self.topwin.attributes("-alpha", 0.8)  # 透明（调试用）
        self.topwin.attributes('-topmost', 1)  # 设置层级最前
        # 创建画布及画布元素。后创建的层级在上。
        self.canvas = tk.Canvas(self.topwin, cursor='plus', bg=None,
                                highlightthickness=0, borderwidth=0)  # 取消边框
        self.canvas.pack(fill='both')
        # 瞄准盒
        rec1 = self.canvas.create_rectangle(  # 实线底层
            self.OB, self.OB, self.OB, self.OB, outline=Config.get('scsColorBoxDown'), width=2)
        rec2 = self.canvas.create_rectangle(  # 虚线表层
            self.OB, self.OB, self.OB, self.OB, outline=Config.get('scsColorBoxUp'), width=2, dash=10)
        self.sightBox = (rec1, rec2)
        self.sightBoxXY = [self.OB, self.OB, self.OB, self.OB]  # 瞄准盒坐标
        # 瞄准线
        lineColor = Config.get('scsColorLine')
        lineW = self.canvas.create_line(  # 纵向
            self.OB, self.OB, self.OB, self.OB, fill=lineColor, width=1)
        lineH = self.canvas.create_line(  # 横向
            self.OB, self.OB, self.OB, self.OB, fill=lineColor, width=1)
        self.sightLine = (lineW, lineH)
        # debug模块
        self.debugXYBox = self.canvas.create_rectangle(  # 坐标下面的底
            self.OB, self.OB, self.OB, self.OB, fill='yellow', outline='#999', width=1)
        self.debugXYText = self.canvas.create_text(self.OB, self.OB,  # 显示坐标
                                                   font=('Microsoft YaHei', 15, 'bold'), fill='red', anchor='nw')
        self.debugList = []  # 显示屏幕信息
        # 闪光模块
        self.flashList = []  # 闪光元素
        # 绑定全局事件
        Hotkey.add('esc', self.__onClose)  # 绑定Esc退出
        Hotkey.add('ctrl+shift+alt+d', self.__switchDebug)  # 切换调试信息
        # 方向键控制鼠标移动
        Hotkey.add('up', lambda: self.__keyMotion(0, -1))
        Hotkey.add('down', lambda: self.__keyMotion(0, 1))
        Hotkey.add('left', lambda: self.__keyMotion(-1, 0))
        Hotkey.add('right', lambda: self.__keyMotion(1, 0))
        # 绑定画布事件
        self.canvas.bind('<Button-1>', self.__onDown)  # 左键按下
        self.canvas.bind('<Button-3>', self.__repaint)  # 右键按下
        self.canvas.bind('<ButtonRelease-1>', self.__onUp)  # 左键松开
        self.canvas.bind('<Motion>', self.__onMotion)  # 鼠标移动
        self.canvas.bind('<Enter>', self.__onMotion)  # 鼠标进入，用于初始化瞄准线
        Log.info('Umi截图启动')

    def __hideElement(self, ele, size=4):  # 隐藏一个画布元素
        # 实际上是挪到画布外
        if size == 2:
            self.canvas.coords(ele, self.OB, self.OB)
        elif size == 4:
            self.canvas.coords(ele, self.OB, self.OB, self.OB, self.OB)

    def __onDown(self, event):  # 鼠标按下
        if self.drawMode == _DrawMode.ready:  # 进入拖拽模式
            self.drawMode = _DrawMode.drag
            # 记录起始点
            self.sightBoxXY[0], self.sightBoxXY[1] = event.x, event.y
            self.sightBoxXY[2], self.sightBoxXY[3] = event.x, event.y
            # 隐藏瞄准线
            for i in (0, 1):
                self.__hideElement(self.sightLine[i], 4)

    def __onUp(self, event):  # 鼠标松开
        if self.drawMode == _DrawMode.drag:  # 离开拖拽模式
            self.drawMode = _DrawMode.ready
            # 记录结束点
            self.sightBoxXY[2], self.sightBoxXY[3] = event.x, event.y
            self.__createGrabImg()  # 生成剪切图像
            self.__onClose()  # 关闭窗口

    def __onMotion(self, event):  # 鼠标移动
        if self.drawMode == _DrawMode.ready:  # 准备模式，刷新瞄准线
            self.canvas.coords(self.sightLine[0],
                               0, event.y, self.scBoxVirtual[4], event.y)
            self.canvas.coords(self.sightLine[1],
                               event.x, 0, event.x, self.scBoxVirtual[5])
        elif self.drawMode == _DrawMode.drag:  # 拖拽模式，刷新瞄准盒
            self.sightBoxXY[2], self.sightBoxXY[3] = event.x, event.y
            for i in (0, 1):
                self.canvas.coords(self.sightBox[i],
                                   self.sightBoxXY[0], self.sightBoxXY[1], event.x, event.y)
        if self.debugList:
            self.canvas.coords(self.debugXYText, event.x+6, event.y+3)
            self.canvas.coords(self.debugXYBox, event.x+3,
                               event.y+3, event.x+130, event.y+28)
            # self.canvas.itemconfig(self.debugXYText, {'text':
            #                                           f'{event.x} , {event.y}'})
            self.canvas.itemconfig(self.debugXYText, {'text':
                                                      f'{event.x_root} , {event.y_root}'})

    def __keyMotion(self, x, y):  # 键盘控制鼠标移动
        if not self.isInitGrab:
            return
        pos = Hotkey.getMousePos()
        pos = (pos[0]+x, pos[1]+y)
        Hotkey.setMousePos(pos)

    def __repaint(self, event):  # 重绘
        Log.info('重绘')
        if self.drawMode == _DrawMode.drag:  # 已在拖拽中
            self.drawMode = _DrawMode.ready  # 退出拖拽模式
            self.sightBoxXY = [self.OB, self.OB, self.OB, self.OB]
            for i in (0, 1):  # 隐藏瞄准盒，显示瞄准线
                self.__hideElement(self.sightBox[i], 4)
            self.canvas.coords(self.sightLine[0],
                               0, event.y, self.scBoxVirtual[4], event.y)
            self.canvas.coords(self.sightLine[1],
                               event.x, 0, event.x, self.scBoxVirtual[5])
        elif self.drawMode == _DrawMode.ready:  # 还在准备中
            self.__onClose()  # 关闭

    def __createGrabImg(self):  # 创建剪切图像
        box = self.sightBoxXY
        if box[0] < 0 and box[1] < 0 and box[2] < 0 and box[3] < 0:
            pass  # 未截图
        elif box[0] == box[2] or box[1] == box[3]:
            pass  # 截图面积为0，无效
        else:
            if box[0] > box[2]:  # 若坐标错位（第二点不在第一点右下角）则交换
                box[0], box[2] = box[2], box[0]
            if box[1] > box[3]:
                box[1], box[3] = box[3], box[1]
            self.sourceBox = tuple(box)  # 记录缩放比例之前的原始box值
            for i in range(4):
                box[i] *= self.allScale  # 乘上缩放比例
            self.imageResult = self.image.crop(box)  # 裁切，产生最终截图数据

    def __onClose(self, event=None):  # 关闭窗口
        if not self.isInitGrab:
            return
        # 隐藏元素
        for i in (0, 1):
            self.__hideElement(self.sightBox[i])
            self.__hideElement(self.sightLine[i])
        self.topwin.withdraw()  # 隐藏窗口
        #  初始化参数
        self.isInitGrab = False
        self.drawMode = _DrawMode.ready
        self.errMsg = None
        flag = self.copyImage()  # 复制图像
        self.image = None  # 删除图像
        self.imageResult = None  # 删除
        _ScreenshotClose(flag, self.errMsg)

    def __flash(self):  # 边缘闪光，提示已截图
        color = 'white'
        width = 100

        def closeFlash():  # 关闭闪光
            for i in self.flashList:
                self.canvas.delete(i)
            self.flashList = []

        for box in self.scBoxList:
            p1x, p1y, p2x, p2y = box
            p1x -= self.scBoxVirtual[0]
            p2x -= self.scBoxVirtual[0]
            p1y -= self.scBoxVirtual[1]
            p2y -= self.scBoxVirtual[1]
            e = self.canvas.create_rectangle(
                p1x, p1y, p2x, p2y, outline=color, width=width)
            self.flashList.append(e)
        self.topwin.after(200, closeFlash)

    def __switchDebug(self, event=None):  # 切换显示/隐藏调试信息
        if not self.isInitGrab:
            return
        color = 'red'
        if self.debugList:  # 删除调试信息
            Config.set('isDebug', False)
            for i in self.debugList:
                self.canvas.delete(i)
            self.debugList = []
            self.__hideElement(self.debugXYBox, 4)
            self.__hideElement(self.debugXYText, 2)
        else:  # 创建调试信息
            Config.set('isDebug', True)
            for index, box in enumerate(self.scBoxList):
                p1x, p1y, p2x, p2y = box
                p1x -= self.scBoxVirtual[0]
                p2x -= self.scBoxVirtual[0]
                p1y -= self.scBoxVirtual[1]
                p2y -= self.scBoxVirtual[1]
                e = self.canvas.create_rectangle(
                    p1x, p1y, p2x, p2y, outline=color, width=3)
                self.debugList.append(e)
                e = self.canvas.create_line(
                    p1x, p1y, p2x, p2y, fill=color, width=3)
                self.debugList.append(e)
                e = self.canvas.create_line(
                    p2x, p1y, p1x, p2y, fill=color, width=3)
                self.debugList.append(e)
                # 文字提示框
                e = self.canvas.create_rectangle(
                    p1x+10, p1y+10, p1x+440, p1y+60, fill='white', width=0)
                self.debugList.append(e)
                e = self.canvas.create_text(p1x+15, p1y+15,
                                            font=('', 15, 'bold'), fill=color, anchor='nw',
                                            text=f'屏幕{index+1}: {box} | {box[2]-box[0]},{box[3]-box[1]}')
                self.debugList.append(e)
                e = self.canvas.create_text(p1x+15, p1y+43,
                                            font=('', 10, ''), fill=color, anchor='nw',
                                            text=f'按 Ctrl+Shift+Alt+D 退出调试模式')
                self.debugList.append(e)
            self.canvas.lift(self.debugXYBox)  # 移动到最上方
            self.canvas.lift(self.debugXYText)  # 移动到最上方

    def copyImage(self):
        '''复制截图到剪贴板。成功返回True，否则False'''
        if not self.imageResult:
            return False
        # 图片转字节
        output = BytesIO()
        self.imageResult.save(output, 'BMP')  # 以位图保存
        imgData = output.getvalue()[14:]  # 去除header
        output.close()
        if Config.get('isShowImage'):  # 显示图片展示窗
            b = self.sourceBox
            p = (b[0], b[1], b[2]-b[0], b[3]-b[1])
            ShowImage(imgPIL=self.imageResult, imgData=imgData, initPos=p)
            return False
        else:  # 直接识别
            try:
                OpenClipboard()  # 打开剪贴板
                EmptyClipboard()  # 清空剪贴板
                SetClipboardData(CF_DIB, imgData)  # 写入
            except Exception as err:
                self.errMsg = f'位图无法写入剪贴板，请检测是否有其他程序正在占用。\n{err}'
                return False
            finally:
                try:
                    CloseClipboard()  # 关闭
                except Exception as err:
                    self.errMsg = f'无法关闭剪贴板。\n{err}'
                    return False
            return True


SSWin = ScreenshotWin()

# class e:
#     def __init__(self, x, y):
#         self.x = x
#         self.y = y
# self.__onDown(e(0, 0))
# self.__onUp(e(50, 20))

# 虚拟屏幕总尺寸  win32api.GetSystemMetrics
# virtualX = GetSystemMetrics(78)  # 常量 win32con.SM_CXVIRTUALSCREEN
# virtualY = GetSystemMetrics(79)  # 常量 win32con.SM_CYVIRTUALSCREEN
# print(f'虚拟尺寸：{virtualX} {virtualY}\n真实尺寸：{self.image.size}')
# print(f'总缩放比例：{self.image.size[0]/virtualX}')



from utils.config import Config, Umi, ScsModeFlag, WindowTopModeFlag  # 最先加载配置
from utils.logger import GetLog
from utils.asset import *  # 资源
from utils.data_structure import KeyList
from utils.tool import Tool
from utils.startup import Startup  # 启动方式
from utils.hotkey import Hotkey  # 快捷键
from utils.command_arg import Parse, Mission  # 启动参数分析
from ui.win_notify import Notify  # 通知弹窗
from ui.win_screenshot import ScreenshotCopy  # 截屏
from ui.win_select_area import IgnoreAreaWin  # 子窗口
from ui.win_ocr_language import ChangeOcrLanguage  # 更改语言
from ui.widget import Widget  # 控件
from ui.pmw.PmwBalloon import Balloon  # 气泡提示
from ui.tray import SysTray
from ocr.engine import OCRe, MsnFlag, EngFlag  # 引擎
# 识图任务处理器
from ocr.msn_batch_paths import MsnBatch
from ocr.msn_quick import MsnQuick

import os
import ctypes
from sys import argv
from PIL import Image  # 图像
import tkinter as tk
import tkinter.font
import tkinter.filedialog
import tkinter.colorchooser
from tkinter import ttk
from windnd import hook_dropfiles  # 文件拖拽
from webbrowser import open as webOpen  # “关于”面板打开项目网址
from argparse import ArgumentParser  # 启动参数

Log = GetLog()


class MainWin:
    def __init__(self):
        self.batList = KeyList()  # 管理批量图片的信息及表格id的列表
        self.tableKeyList = []  # 顺序存放self.imgDict
        self.lockWidget = []  # 需要运行时锁定的组件

        # 1.初始化主窗口
        self.win = tk.Tk()
        self.win.withdraw()  # 隐藏窗口，等初始化完毕再考虑是否显示
        self.balloon = Balloon(self.win)  # 气泡框

        def initStyle():  # 初始化样式
            style = ttk.Style()
            # winnative clam alt default classic vista xpnative
            # style.theme_use('default')
            style.configure('icon.TButton', padding=(12, 0))
            style.configure('go.TButton', font=('Microsoft YaHei', '12', ''),  # bold
                            width=9)
            style.configure('gray.TCheckbutton', foreground='gray')
        initStyle()

        def initDPI():
            # 调用api设置成由应用程序缩放
            ctypes.windll.shcore.SetProcessDpiAwareness(1)
            # 调用api获得当前的缩放因子
            ScaleFactor = ctypes.windll.shcore.GetScaleFactorForDevice(0)
            # 设置缩放因子
            self.win.tk.call('tk', 'scaling', ScaleFactor/100)
        # initDPI()

        def initWin():
            self.win.title(Umi.name)
            # 窗口大小与位置
            w, h = 360, 500  # 窗口初始大小与最小大小
            ws, hs = self.win.winfo_screenwidth(), self.win.winfo_screenheight()
            x, y = round(ws/2 - w/2), round(hs/2 - h/2)  # 初始位置，屏幕正中
            self.win.minsize(w, h)  # 最小大小
            self.win.geometry(f"{w}x{h}+{x}+{y}")  # 初始大小与位置
            self.win.protocol("WM_DELETE_WINDOW", self.onClose)  # 窗口关闭
            # 注册文件拖入，整个主窗口内有效
            # 改成延迟一段时间后生效，减少产生异常的概率
            # Fatal Python error: PyEval_RestoreThread: NULL tstate
            # hook_dropfiles(self.win, func=self.draggedImages)
            hook_dropfiles(self.win, func=lambda e: self.win.after(
                80, lambda: self.draggedImages(e)))
            # 图标
            Asset.initRelease()  # 释放base64资源到本地
            Asset.initTK()  # 初始化tk图片
            self.win.iconphoto(False, Asset.getImgTK('umiocr24'))  # 设置窗口图标
        initWin()

        # 2.初始化配置项
        self.win.bind('<<QuitEvent>>', lambda *e: self.onClose())  # 退出事件
        Config.initTK(self)  # 初始化设置项
        Config.load()  # 加载本地文件
        Config.checkMultiOpen()  # 检查多开

        # 3.初始化组件
        def initTop():  # 顶部按钮
            tk.Frame(self.win, height=5).pack(side='top')
            fr = tk.Frame(self.win)
            fr.pack(side='top', fill="x", padx=5)
            # 右侧按钮
            self.btnRun = ttk.Button(fr, command=self.run, text='开始任务',
                                     style='go.TButton')
            self.btnRun.pack(side='right', fill='y')
            # 左侧文本和进度条
            vFrame2 = tk.Frame(fr)
            vFrame2.pack(side='top', fill='x')
            # 进度条上方
            wid = ttk.Checkbutton(vFrame2, variable=Config.getTK('isWindowTop'),
                                  text='窗口置顶', style='gray.TCheckbutton')
            wid.pack(side='left')
            self.balloon.bind(
                wid, '窗口锁定于系统顶层\n\n启用后，软件内的鼠标悬停提示框会被隐藏')
            tk.Label(vFrame2, textvariable=Config.getTK('tipsTop2')).pack(
                side='right', padx=2)
            tk.Label(vFrame2, textvariable=Config.getTK('tipsTop1')).pack(
                side='right', padx=2)
            self.progressbar = ttk.Progressbar(fr)
            self.progressbar.pack(side='top', padx=2, pady=2, fill="x")
        initTop()

        self.notebook = ttk.Notebook(self.win)  # 初始化选项卡组件
        self.notebook.pack(expand=True, fill=tk.BOTH)  # 填满父组件
        self.notebookTab = []

        def initTab1():  # 表格卡
            tabFrameTable = tk.Frame(self.notebook)  # 选项卡主容器
            self.notebookTab.append(tabFrameTable)
            self.notebook.add(tabFrameTable, text=f'{"批量处理": ^10s}')
            # 顶栏
            fr1 = tk.Frame(tabFrameTable)
            fr1.pack(side='top', fill='x', padx=1, pady=1)
            # 左
            btn = ttk.Button(fr1, image=Asset.getImgTK('screenshot24'),  # 截图按钮
                             command=self.openScreenshot,
                             style='icon.TButton',  takefocus=0,)
            self.balloon.bind(
                btn, '屏幕截图　　说明\n左键拖拽：框选区域\n右键点击：取消框选\n　　 Esc：退出截图')
            btn.pack(side='left')
            self.lockWidget.append(btn)
            btn = ttk.Button(fr1, image=Asset.getImgTK('paste24'),  # 剪贴板按钮
                             command=self.runClipboard,
                             style='icon.TButton',  takefocus=0,)
            self.balloon.bind(btn, '粘贴图片')
            btn.pack(side='left')
            self.lockWidget.append(btn)
            btn = ttk.Button(fr1, image=Asset.getImgTK('language24'),  # 语言按钮
                             command=ChangeOcrLanguage,
                             style='icon.TButton',  takefocus=0)
            self.balloon.bind(btn, '更改OCR语言')
            btn.pack(side='left')
            self.lockWidget.append(btn)
            # 右
            btn = ttk.Button(fr1, image=Asset.getImgTK('clear24'),  # 清空按钮
                             command=self.clearTable,
                             style='icon.TButton',  takefocus=0,)
            self.balloon.bind(btn, '清空表格')
            btn.pack(side='right')
            self.lockWidget.append(btn)
            btn = ttk.Button(fr1, image=Asset.getImgTK('delete24'),  # 删除按钮
                             command=self.delImgList,
                             style='icon.TButton',  takefocus=0,)
            self.balloon.bind(btn, '移除选中的文件\n按住Shift或Ctrl，点击左键可选中多个文件')
            btn.pack(side='right')
            self.lockWidget.append(btn)
            btn = ttk.Button(fr1, image=Asset.getImgTK('openfile24'),  # 打开文件按钮
                             command=self.openFileWin,
                             style='icon.TButton',  takefocus=0,)
            self.balloon.bind(btn, '浏览文件')
            btn.pack(side='right')
            self.lockWidget.append(btn)
            # 表格主体
            fr2 = tk.Frame(tabFrameTable)
            fr2.pack(side='top', fill='both')
            self.table = ttk.Treeview(
                master=fr2,  # 父容器
                height=50,  # 表格显示的行数,height行
                columns=['name', 'time', 'score'],  # 显示的列
                show='headings',  # 隐藏首列
            )
            self.table.pack(expand=True, side="left", fill='both')
            self.table.heading('name', text='文件名称')
            self.table.heading('time', text='耗时')
            self.table.heading('score', text='置信度')
            self.table.column('name', minwidth=40)
            self.table.column('time', width=20, minwidth=20)
            self.table.column('score', width=30, minwidth=30)
            vbar = tk.Scrollbar(  # 绑定滚动条
                fr2, orient='vertical', command=self.table.yview)
            vbar.pack(side="left", fill='y')
            self.table["yscrollcommand"] = vbar.set
        initTab1()

        def initTab2():  # 输出卡
            tabFrameOutput = tk.Frame(self.notebook)  # 选项卡主容器
            self.notebookTab.append(tabFrameOutput)
            self.notebook.add(tabFrameOutput, text=f'{"识别内容": ^10s}')
            fr1 = tk.Frame(tabFrameOutput)
            fr1.pack(side='top', fill='x', padx=1, pady=1)
            self.isAutoRoll = tk.IntVar()
            self.isAutoRoll.set(1)
            # 左
            btn = ttk.Button(fr1, image=Asset.getImgTK('screenshot24'),  # 截图按钮
                             command=self.openScreenshot,
                             style='icon.TButton',  takefocus=0,)
            self.balloon.bind(
                btn, '屏幕截图　　说明\n左键拖拽：框选区域\n右键点击：取消框选\n　　 Esc：退出截图')
            btn.pack(side='left')
            self.lockWidget.append(btn)
            btn = ttk.Button(fr1, image=Asset.getImgTK('paste24'),  # 剪贴板按钮
                             command=self.runClipboard,
                             style='icon.TButton',  takefocus=0,)
            self.balloon.bind(btn, '粘贴图片')
            btn.pack(side='left')
            self.lockWidget.append(btn)
            btn = ttk.Button(fr1, image=Asset.getImgTK('language24'),  # 语言按钮
                             command=ChangeOcrLanguage,
                             style='icon.TButton',  takefocus=0)
            self.balloon.bind(btn, '更改OCR语言')
            btn.pack(side='left')
            self.lockWidget.append(btn)

            # 右
            btn = ttk.Button(fr1, image=Asset.getImgTK('clear24'),  # 清空按钮
                             command=self.panelClear,
                             style='icon.TButton',  takefocus=0,)
            self.balloon.bind(btn, '清空输出面板\n在【设置→快捷识图】可启用自动清空面板')
            btn.pack(side='right')

            ttk.Checkbutton(fr1, variable=self.isAutoRoll, text="自动滚动",
                            takefocus=0,).pack(side='right')
            tf = tk.Label(fr1, text='字体', fg='gray', cursor='hand2')
            tf.pack(side='right', padx=10)
            tf.bind(
                '<Button-1>', lambda *e: self.notebook.select(self.notebookTab[2]))  # 转到设置卡
            self.balloon.bind(tf, '在【设置】选项卡更改输出面板的字体')

            fr2 = tk.Frame(tabFrameOutput)
            fr2.pack(side='top', fill='both')
            vbar = tk.Scrollbar(fr2, orient='vertical')  # 滚动条
            vbar.pack(side="right", fill='y')
            self.textOutput = tk.Text(fr2, height=500, width=500)
            self.textOutput.pack(fill='both', side="left")
            self.textOutput.tag_config(  # 添加高亮标签
                'blue', foreground='blue')
            self.textOutput.tag_config(  # 添加高亮标签
                'red', foreground='red')
            vbar["command"] = self.textOutput.yview
            self.textOutput["yscrollcommand"] = vbar.set
        initTab2()

        def initTab3():  # 设置卡
            tabFrameConfig = tk.Frame(self.notebook)  # 选项卡主容器
            self.notebookTab.append(tabFrameConfig)
            self.notebook.add(tabFrameConfig, text=f'{"设置": ^10s}')

            def initOptFrame():  # 初始化可滚动画布 及 内嵌框架
                optVbar = tk.Scrollbar(
                    tabFrameConfig, orient="vertical")  # 创建滚动条
                optVbar.pack(side="right", fill="y")
                self.optCanvas = tk.Canvas(
                    tabFrameConfig, highlightthickness=0)  # 创建画布，用于承载框架。highlightthickness取消高亮边框
                self.optCanvas.pack(side="left", fill="both",
                                    expand="yes")  # 填满父窗口
                self.optCanvas["yscrollcommand"] = optVbar.set  # 绑定滚动条
                optVbar["command"] = self.optCanvas.yview
                self.optFrame = tk.Frame(self.optCanvas)  # 容纳设置项的框架
                self.optFrame.pack()
                self.optCanvas.create_window(  # 框架塞进画布
                    (0, 0), window=self.optFrame, anchor="nw")
            initOptFrame()

            LabelFramePadY = 3  # 每个区域上下间距

            def initTopTips():  # 顶部提示
                fTips = tk.Frame(self.optFrame)
                fTips.pack(side='top')
                tipsLab = tk.Label(
                    fTips, fg='red',
                    text='关闭窗口置顶，方可显示鼠标悬停提示框')
                if Config.get('isWindowTop'):
                    tipsLab.pack(side='top')
                tk.Frame(fTips).pack(side='top')  # 空框架，用于自动调整高度的占位

                def changeIsWinTop():
                    if Config.get('isWindowTop'):  # 启用置顶
                        tipsLab.pack(side='top')
                    else:  # 取消置顶
                        tipsLab.pack_forget()
                    self.gotoTop()
                Config.addTrace('isWindowTop', changeIsWinTop)

            initTopTips()

            def initSoftwareFrame():  # 软件行为设置
                fSoft = tk.LabelFrame(
                    self.optFrame, text='通用设置')
                fSoft.pack(side='top', fill='x',
                           ipady=2, pady=LabelFramePadY, padx=4)

                # 主面板字体设置
                fr3 = tk.Frame(fSoft)
                fr3.pack(side='top', fill='x', pady=2, padx=5)
                fr3.grid_columnconfigure(1, weight=1)
                self.balloon.bind(fr3, '调整【识别内容】选项卡中输出面板的字体样式')
                tk.Label(fr3, text='输出面板字体').grid(column=0, row=0, sticky='w')
                ff = tk.font.families()  # 获取系统字体
                fontFamilies = []
                fontFamiliesABC = []
                for i in ff:
                    if not i[0] == '@':  # 排除竖版
                        if '\u4e00' <= i[0] <= '\u9fff':  # 中文开头的优先
                            fontFamilies.append(i)
                        else:
                            fontFamiliesABC.append(i)
                fontFamilies += fontFamiliesABC
                cbox = ttk.Combobox(fr3, state='readonly', takefocus=0,
                                    textvariable=Config.getTK('textpanelFontFamily'), value=fontFamilies)
                cbox.grid(column=1, row=0, sticky='ew')
                self.balloon.bind(cbox, '不要使用滚轮。\n请用上下方向键或拉动滚动条来浏览列表')
                tk.Label(fr3, text='字号').grid(column=2, row=0, sticky='w')
                tk.Entry(fr3, textvariable=Config.getTK('textpanelFontSize'),
                         width=4, takefocus=0).grid(column=3, row=0, sticky='w')
                tk.Label(fr3, text=' ').grid(column=4, row=0, sticky='w')
                ttk.Checkbutton(fr3, text='加粗',
                                variable=Config.getTK('isTextpanelFontBold')).grid(column=5, row=0, sticky='w')
                # 检查当前配置字体是否存在
                f = Config.get('textpanelFontFamily')
                if f and f not in fontFamilies:
                    Log.error(f'配置输出面板字体【{f}】不存在。重置为空')
                    Config.set('textpanelFontFamily', '')

                def updateTextpanel():
                    f = Config.get('textpanelFontFamily')
                    s = Config.get('textpanelFontSize')
                    b = Config.get('isTextpanelFontBold')
                    font = (f, s, 'bold' if b else 'normal')
                    self.textOutput['font'] = font
                Config.addTrace('textpanelFontFamily', updateTextpanel)
                Config.addTrace('textpanelFontSize', updateTextpanel)
                Config.addTrace('isTextpanelFontBold', updateTextpanel)
                updateTextpanel()

                fr1 = tk.Frame(fSoft)
                fr1.pack(side='top', fill='x', pady=2, padx=5)
                fr1.grid_columnconfigure(1, weight=1)
                self.balloon.bind(
                    fr1, '可关闭/开启系统托盘图标，修改双击图标时触发的功能\n该项目修改后，下次打开软件生效')
                wid = ttk.Checkbutton(fr1, text='显示系统托盘图标',
                                      variable=Config.getTK('isTray'))
                wid.grid(column=0, row=0, sticky='w')
                Widget.comboboxFrame(fr1, '，双击图标', 'clickTrayMode', width=12).grid(
                    column=1, row=0, sticky='w')

                fr2 = tk.Frame(fSoft)
                fr2.pack(side='top', fill='x', pady=2, padx=5)
                tk.Label(fr2, text='主窗口关闭：').pack(side='left', padx=2)
                ttk.Radiobutton(fr2, text='退出软件',
                                variable=Config.getTK('isBackground'), value=False).pack(side='left')
                wid = ttk.Radiobutton(fr2, text='最小化到托盘',
                                      variable=Config.getTK('isBackground'), value=True)
                wid.pack(side='left', padx=15)
                self.balloon.bind(wid, '显示系统托盘图标时，本选项才有效')

                # 弹出方式设置
                fr3 = tk.Frame(fSoft)
                fr3.pack(side='top', fill='x', pady=2, padx=5)
                tk.Label(fr3, text='主窗口弹出：').pack(side='left', padx=2)
                wid = ttk.Radiobutton(fr3, text='自动弹出',
                                      variable=Config.getTK('WindowTopMode'), value=WindowTopModeFlag.finish)
                wid.pack(side='left')
                self.balloon.bind(
                    wid, '唤起快捷识图、或批量任务完成时弹出主窗口')
                wid = ttk.Radiobutton(fr3, text='静默模式',
                                      variable=Config.getTK('WindowTopMode'), value=WindowTopModeFlag.never)
                wid.pack(side='left', padx=15)
                self.balloon.bind(
                    wid, '不会主动弹出窗口\n建议启用通知弹窗')

                # 消息弹窗设置
                def changeNotify():
                    if Config.get('isNotify'):
                        Notify('欢迎使用 Umi-OCR', '通知弹窗已开启')
                Config.addTrace('isNotify', changeNotify)
                fr4 = tk.Frame(fSoft)
                fr4.pack(side='top', fill='x', pady=2, padx=5)
                ttk.Checkbutton(
                    fr4, variable=Config.getTK('isNotify'), text='启用通知弹窗').pack(side='left')

                # 启动方式设置
                fr5 = tk.Frame(fSoft)
                fr5.pack(side='top', fill='x', pady=2, padx=5)
                self.balloon.bind(
                    fr5, '可设置静默启动，收纳到系统托盘，不显示主窗口')
                ttk.Checkbutton(fr5, variable=Config.getTK('isAutoStartup'),
                                text='开机自启', command=Startup.switchAutoStartup).pack(side='left')
                ttk.Checkbutton(fr5, variable=Config.getTK('isStartMenu'),
                                text='开始菜单项', command=Startup.switchStartMenu).pack(side='left', padx=20)
                ttk.Checkbutton(fr5, variable=Config.getTK('isDesktop'),
                                text='桌面快捷方式', command=Startup.switchDesktop).pack(side='left')
            initSoftwareFrame()

            def quickOCR():  # 快捷识图设置
                fQuick = tk.LabelFrame(
                    self.optFrame, text='快捷识图')
                fQuick.pack(side='top', fill='x',
                            ipady=2, pady=LabelFramePadY, padx=4)
                # 截图快捷键触发时，子线程向主线程发送事件，在主线程中启动截图窗口
                # 避免子线程直接唤起截图窗导致的窗口闪烁现象
                self.win.bind('<<ScreenshotEvent>>',
                              self.openScreenshot)  # 绑定截图事件
                cbox = Widget.comboboxFrame(fQuick, '截图模块：', 'scsMode')
                cbox.pack(side='top', fill='x', padx=4)
                self.balloon.bind(
                    cbox, '切换截图工作模块\n\n【Umi-OCR 软件截图】方便、精准\n【Windows 系统截图】兼容性更好')
                frss = tk.Frame(fQuick)
                frss.pack(side='top', fill='x')
                fhkUmi = tk.Frame(frss)
                fhkUmi.pack(side='top', fill='x')
                fhkU0 = tk.Frame(fhkUmi)
                fhkU0.pack(side='top', fill='x', pady=2)
                tk.Label(fhkU0, text='指示器颜色：').pack(side='left')
                self.balloon.bind(fhkU0, '修改截图时指示器的颜色\n该项目修改后，下次打开软件生效')

                def changeColor(configName, title=None):
                    initColor = Config.get(configName)
                    color = tk.colorchooser.askcolor(
                        color=initColor, title=title)
                    if color[1]:
                        Config.set(configName, color[1])
                lab1 = tk.Label(fhkU0, text='十字线', cursor='hand2', fg='blue')
                lab1.pack(side='left', padx=9)
                lab1.bind(
                    '<Button-1>', lambda *e: changeColor('scsColorLine', '截图十字线颜色'))
                lab2 = tk.Label(fhkU0, text='虚线框表层', cursor='hand2', fg='blue')
                lab2.pack(side='left', padx=9)
                lab2.bind(
                    '<Button-1>', lambda *e: changeColor('scsColorBoxUp', '截图矩形框 虚线表层颜色'))
                lab3 = tk.Label(fhkU0, text='虚线框底层', cursor='hand2', fg='blue')
                lab3.pack(side='left', padx=9)
                lab3.bind(
                    '<Button-1>', lambda *e: changeColor('scsColorBoxDown', '截图矩形框 虚线底层颜色'))
                wid = Widget.hotkeyFrame(fhkUmi, '截图识别　快捷键　', 'Screenshot',
                                         lambda *e: self.win.event_generate(
                                             '<<ScreenshotEvent>>'), isAutoBind=False)
                wid.pack(side='top', fill='x')
                self.balloon.bind(
                    wid, '关闭快捷键后，仍能通过面板上的按钮或托盘小图标调用截图\n点击【修改】可设置自定义快捷键')

                syssscom = 'win+shift+s'
                fhkSys = Widget.hotkeyFrame(frss, '系统截图　快捷键　', 'Screenshot',
                                            lambda *e: self.win.event_generate(
                                                '<<ScreenshotEvent>>'), True, syssscom, isAutoBind=False)
                self.balloon.bind(
                    fhkSys, '监听到系统截图后调用OCR\n\n若截图后软件没有反应，请确保windows系统自带的\n【截图和草图】中【自动复制到剪贴板】开关处于打开状态')

                wid = Widget.hotkeyFrame(
                    fQuick, '粘贴图片　快捷键　', 'Clipboard', self.runClipboard, isAutoBind=True)
                wid.pack(side='top', fill='x', padx=4)
                self.balloon.bind(wid, '尝试读取剪贴板，若存在图片则调用OCR\n点击【修改】可设置自定义快捷键')
                if Config.get('isAdvanced'):  # 隐藏高级选项：组合键判定调节
                    fr1 = tk.Frame(fQuick)
                    fr1.pack(side='top', fill='x', pady=2, padx=5)
                    tk.Label(fr1, text='　 组合键：').pack(side='left')
                    fr11 = tk.Frame(fr1)
                    fr11.pack(side='left')
                    self.balloon.bind(
                        fr11, '宽松：当前按下的按键只要包含设定的组合键，就能触发\n严格：当前按下的按键必须与设定的组合一致，才能触发')
                    tk.Label(fr11, text='触发判定').pack(side='left')
                    ttk.Radiobutton(fr11, text='宽松',
                                    variable=Config.getTK('isHotkeyStrict'), value=False).pack(side='left')
                    ttk.Radiobutton(fr11, text='严格',
                                    variable=Config.getTK('isHotkeyStrict'), value=True).pack(side='left')
                    fr12 = tk.Frame(fr1)
                    fr12.pack(side='left')
                    self.balloon.bind(fr12, '必须在该时间之内\n连续按下组合中的所有按键，才能触发')
                    tk.Label(fr12, text='，时限').pack(side='left')
                    tk.Entry(fr12,
                             textvariable=Config.getTK('hotkeyMaxTtl'), width=4).pack(side='left')
                    tk.Label(fr12, text='秒').pack(side='left')

                fr2 = tk.Frame(fQuick)
                fr2.pack(side='top', fill='x', pady=2, padx=5)
                fr2.grid_columnconfigure(1, minsize=20)
                wid = ttk.Checkbutton(fr2, variable=Config.getTK('isScreenshotHideWindow'),
                                      text='隐藏主窗口')
                wid.grid(column=0, row=0, sticky='w')
                self.balloon.bind(
                    wid, f'截图前隐藏主窗口\n会延迟{Config.get("screenshotHideWindowWaitTime")}毫秒以等待窗口动画')
                wid = ttk.Checkbutton(fr2, variable=Config.getTK('isShowImage'),
                                      text='截图预览窗口')
                wid.grid(column=2, row=0)
                self.balloon.bind(
                    wid, f'不勾选：截图后立刻OCR识图\n勾选：截图后展示图片，可稍后识别或保存图片')
                wid = ttk.Checkbutton(fr2, variable=Config.getTK('isNeedCopy'),
                                      text='自动复制结果')
                wid.grid(column=0, row=1)
                self.balloon.bind(wid, '快捷识图完成后，将得到的文本复制到剪贴板')
                wid = ttk.Checkbutton(fr2, variable=Config.getTK('isNeedClear'),
                                      text='自动清空面板')
                wid.grid(column=2, row=1)
                self.balloon.bind(wid, f'每次快捷识图将清空识别内容面板，同时省略时间等信息')

                if Config.get('isAdvanced'):  # 隐藏高级选项：截图联动
                    frSend = tk.Frame(fQuick)
                    frSend.pack(side='top', fill='x', pady=2, padx=4)
                    frSend.grid_columnconfigure(0, weight=1)
                    self.balloon.bind(frSend, '截图联动：按下快捷键，执行截图OCR并将结果复制到剪贴板，\n然后发送指定键盘按键\n可用于联动唤起翻译器或AHK等工具\n次：重复发送按键的次数，如2为双击')
                    wid = Widget.hotkeyFrame(
                        frSend, '截图联动　快捷键　', 'FinishSend', func=self.openLinkageScreenshot, isAutoBind=True)
                    wid.grid(column=0, row=0, sticky="nsew")
                    wid = Widget.hotkeyFrame(
                        frSend, '　 联动发送按键　', 'FinishSend2', isAutoBind=False, isCheckBtn=False)
                    wid.grid(column=0, row=1, sticky="nsew")
                    tk.Entry(frSend, width=2, textvariable=Config.getTK('hotkeyFinishSendNumber')
                            ).grid(column=1, row=1)
                    tk.Label(frSend, text='次').grid(column=2, row=1)

                # 切换截图模式
                def onModeChange():
                    isHotkey = Config.get('isHotkeyScreenshot')
                    scsName = Config.get('scsModeName')
                    umihk = Config.get('hotkeyScreenshot')
                    scsMode = Config.get('scsMode').get(
                        scsName, ScsModeFlag.multi)  # 当前截屏模式
                    if scsMode == ScsModeFlag.system:  # 切换到系统截图
                        fhkUmi.forget()
                        fhkSys.pack(side='top', fill='x', padx=4)
                        self.updateFrameHeight()  # 刷新框架
                        if isHotkey:  # 当前已在注册
                            if umihk:  # 注销软件截图
                                Widget.delHotkey(umihk)  # 注销按键
                            Hotkey.add(syssscom,  # 添加快捷键监听
                                       lambda *e: self.win.event_generate('<<ScreenshotEvent>>'))
                    elif scsMode == ScsModeFlag.multi:  # 切换到软件截图
                        fhkSys.forget()
                        fhkUmi.pack(side='top', fill='x', padx=4)
                        self.updateFrameHeight()  # 刷新框架
                        if isHotkey:
                            Widget.delHotkey(syssscom)  # 注销按键
                            if umihk:
                                Hotkey.add(umihk,  # 添加快捷键监听
                                           lambda *e: self.win.event_generate('<<ScreenshotEvent>>'))
                    Log.info(f'截图模式改变：{scsMode}')
                Config.addTrace('scsModeName', onModeChange)
                onModeChange()
            quickOCR()

            # 批量任务设置
            frameBatch = tk.LabelFrame(self.optFrame, text="批量任务")
            frameBatch.pack(side='top', fill='x',
                            ipady=2, pady=LabelFramePadY, padx=4)

            def initScheduler():  # 计划任务设置
                frameScheduler = tk.LabelFrame(
                    frameBatch, labelanchor='n', text="计划任务")
                frameScheduler.pack(side='top', fill='x',
                                    ipady=2, pady=LabelFramePadY, padx=4)

                fr1 = tk.Frame(frameScheduler)
                fr1.pack(side='top', fill='x', pady=2, padx=5)
                ttk.Checkbutton(fr1, text="完成后打开文件",
                                variable=Config.getTK('isOpenOutputFile')).pack(side='left')
                ttk.Checkbutton(fr1, text="完成后打开目录",
                                variable=Config.getTK('isOpenExplorer'),).pack(side='left', padx=15)

                fr2 = tk.Frame(frameScheduler)
                fr2.pack(side='top', fill='x', pady=2, padx=5)
                ttk.Checkbutton(fr2, text='本次完成后执行',
                                variable=Config.getTK('isOkMission')).pack(side='left')
                okMissionDict = Config.get("okMission")
                okMissionNameList = [i for i in okMissionDict.keys()]
                wid = ttk.Combobox(fr2, width=14, state="readonly", textvariable=Config.getTK('okMissionName'),
                                   value=okMissionNameList)
                wid.pack(side='left')
                self.balloon.bind(wid, '可打开软件配置json文件，添加自己的任务（cmd命令）')
                if Config.get("okMissionName") not in okMissionNameList:
                    wid.current(0)  # 初始化Combobox和okMissionName
            initScheduler()

            def initInFile():  # 输入设置
                fInput = tk.LabelFrame(
                    frameBatch, labelanchor='n', text='图片导入')
                fInput.pack(side='top', fill='x',
                                 ipady=2, pady=LabelFramePadY, padx=4)
                self.balloon.bind(
                    fInput, f"允许的图片格式：\n{Config.get('imageSuffix')}")

                fr1 = tk.Frame(fInput)
                fr1.pack(side='top', fill='x', pady=2, padx=5)
                wid = ttk.Checkbutton(
                    fr1, variable=Config.getTK('isRecursiveSearch'), text='递归读取子文件夹中所有图片')
                wid.grid(column=0, row=0, columnspan=2, sticky='w')
                self.lockWidget.append(wid)
                if Config.get('isAdvanced'):  # 隐藏高级选项：修改图片许可后缀
                    tk.Label(fr1, text='图片后缀：　').grid(
                        column=0, row=2, sticky='w')
                    enInSuffix = tk.Entry(
                        fr1, textvariable=Config.getTK('imageSuffix'))
                    enInSuffix.grid(column=1, row=2, sticky='nsew')
                    self.lockWidget.append(enInSuffix)

                fr1.grid_columnconfigure(1, weight=1)
            initInFile()

            def initOutFile():  # 输出设置
                fOutput = tk.LabelFrame(
                    frameBatch, labelanchor='n', text="结果输出")
                fOutput.pack(side='top', fill='x',
                                  ipady=2, pady=LabelFramePadY, padx=4)
                # 输出文件类型勾选
                fr1 = tk.Frame(fOutput)
                fr1.pack(side='top', fill='x', pady=2, padx=5)

                wid = ttk.Checkbutton(
                    fr1, variable=Config.getTK('isOutputTxt'), text='合并.txt文件')
                self.balloon.bind(wid, f'所有识别文本输出到同一个txt文件')
                wid.grid(column=0, row=0,  sticky='w')
                self.lockWidget.append(wid)
                wid = ttk.Checkbutton(
                    fr1, variable=Config.getTK('isOutputSeparateTxt'), text='独立.txt文件')
                self.balloon.bind(wid, f'每张图片的文本输出到同名的单独txt文件')
                wid.grid(column=2, row=0,  sticky='w')
                self.lockWidget.append(wid)
                wid = ttk.Checkbutton(
                    fr1, variable=Config.getTK('isOutputMD'), text='图文链接.md文件')
                self.balloon.bind(wid, f'使用Markdown阅读器打开，可同时展示图片和文本')
                wid.grid(column=0, row=1,  sticky='w')
                self.lockWidget.append(wid)
                wid = ttk.Checkbutton(
                    fr1, variable=Config.getTK('isOutputJsonl'), text='原始信息.jsonl文件')
                self.balloon.bind(wid, f'含所有文件路径和OCR信息，可导入别的程序进一步操作')
                wid.grid(column=2, row=1,  sticky='w')
                self.lockWidget.append(wid)
                tk.Label(fr1, text=' ').grid(column=1, row=0)

                def offAllOutput(e):  # 关闭全部输出
                    if OCRe.msnFlag == MsnFlag.none:
                        Config.set('isOutputTxt', False)
                        Config.set('isOutputSeparateTxt', False)
                        Config.set('isOutputMD', False)
                        Config.set('isOutputJsonl', False)
                labelOff = tk.Label(fr1, text='　 关闭全部输出',
                                    cursor='hand2', fg='blue')
                labelOff.grid(column=0, row=2, sticky='w')
                labelOff.bind('<Button-1>', offAllOutput)  # 绑定关闭全部输出

                wid = ttk.Checkbutton(fr1, text='图片中不含文字时，不输出信息',
                                      variable=Config.getTK('isIgnoreNoText'),)
                wid.grid(column=0, row=10, columnspan=9, sticky='w')
                self.lockWidget.append(wid)

                tk.Label(fOutput, fg='gray',
                         text="下面两项为空时，默认输出到第一张图片所在的文件夹"
                         ).pack(side='top', fill='x', padx=5)
                # 输出目录
                fr2 = tk.Frame(fOutput)
                fr2.pack(side='top', fill='x', pady=2, padx=5)
                tk.Label(fr2, text="输出目录：").grid(column=0, row=3, sticky='w')
                enOutPath = tk.Entry(
                    fr2, textvariable=Config.getTK('outputFilePath'))
                enOutPath.grid(column=1, row=3,  sticky='ew')
                self.lockWidget.append(enOutPath)
                fr2.grid_rowconfigure(4, minsize=2)  # 第二行拉开间距
                tk.Label(fr2, text="输出文件名：").grid(column=0, row=5, sticky='w')
                enOutName = tk.Entry(
                    fr2, textvariable=Config.getTK('outputFileName'))
                enOutName.grid(column=1, row=5, sticky='ew')
                self.lockWidget.append(enOutName)
                fr2.grid_columnconfigure(1, weight=1)  # 第二列自动扩充
            initOutFile()

            # 后处理设置
            def initProcess():  # 后处理设置
                fProcess = tk.LabelFrame(self.optFrame,  text='文本后处理')
                fProcess.pack(side='top', fill='x',
                              ipady=2, pady=LabelFramePadY, padx=4)

                fIgnore = tk.Frame(fProcess)
                fIgnore.pack(side='top', fill='x', pady=2, padx=4)

                self.ignoreBtn = ttk.Button(fIgnore, text='打开忽略区域编辑器（设置排除水印）',
                                            command=self.openSelectArea)
                self.ignoreBtn.pack(side='top', fill='x')
                self.balloon.bind(
                    self.ignoreBtn, '忽略图片中的指定区域\n可用于批量识别时排除图片水印')
                self.lockWidget.append(self.ignoreBtn)
                # 忽略区域本体框架
                self.ignoreFrame = tk.Frame(fIgnore)  # 不pack，动态添加
                self.ignoreFrame.grid_columnconfigure(0, minsize=4)
                wid = ttk.Button(self.ignoreFrame, text='添加区域',
                                 command=self.openSelectArea)
                wid.grid(column=1, row=0, sticky='w')
                self.lockWidget.append(wid)
                wid = ttk.Button(self.ignoreFrame, text='清空区域',
                                 command=self.clearArea)
                wid.grid(column=1, row=1, sticky='w')
                self.lockWidget.append(wid)
                self.ignoreLabel = tk.Label(
                    self.ignoreFrame, anchor='w', justify='left')  # 显示生效大小
                self.ignoreLabel.grid(column=1, row=2, sticky='n')
                self.balloon.bind(
                    self.ignoreLabel, '批量任务时，只有分辨率与之相同的图片，才会应用忽略区域。')
                self.ignoreFrame.grid_rowconfigure(2, minsize=10)
                self.ignoreFrame.grid_columnconfigure(2, minsize=4)
                self.canvasHeight = 120  # 画板高度不变，宽度根据选区回传数据调整
                self.canvas = tk.Canvas(self.ignoreFrame, width=200, height=self.canvasHeight,
                                        bg="black", cursor='hand2')
                self.canvas.grid(column=3, row=0, rowspan=10)
                self.canvas.bind(
                    '<Button-1>', lambda *e: self.openSelectArea())
                fpro = tk.Frame(fProcess)
                fpro.pack(side='top', fill='x', pady=2, padx=4)
                fpro.grid_columnconfigure(0, weight=1)
                wid = Widget.comboboxFrame(
                    fpro, '合并段落：　', 'tbpu', self.lockWidget)
                wid.grid(column=0, row=0, sticky='ew')
                self.balloon.bind(wid, '将OCR划分的单行文本合并成整段文字\n点击右侧按钮，浏览方案说明')
                labelUse = tk.Label(fpro, text='说明', width=5,
                                    fg='deeppink', cursor='question_arrow')
                labelUse.grid(column=1, row=0)
                labelUse.bind(
                    '<Button-1>', lambda *e: self.showTips(GetTbpuHelp(Umi.website)))  # 绑定鼠标左键点击
            initProcess()

            def initOcrUI():  # OCR引擎设置
                frameOCR = tk.LabelFrame(
                    self.optFrame, text="OCR识别引擎设置")
                frameOCR.pack(side='top', fill='x', ipady=2,
                              pady=LabelFramePadY, padx=4)
                wid = Widget.comboboxFrame(
                    frameOCR, '识别语言：　', 'ocrConfig', self.lockWidget)
                wid.pack(side='top', fill='x', pady=2, padx=5)
                self.balloon.bind(
                    wid, '本软件有整理好的多国语言扩展包，可导入更多语言模型库，\n也可以手动导入PaddleOCR兼容的模型库，\n详情请浏览项目Github主页\n\n竖排模型库（识别语言）建议与竖排合并段落搭配使用')
                # 压缩
                fLim = tk.Frame(frameOCR)
                fLim.pack(side='top', fill='x', pady=2, padx=5)
                self.balloon.bind(
                    fLim, '长边压缩模式可以大幅加快识别速度，但可能降低大分辨率图片的识别准确率\n大于4000像素的图片，可将数值改为最大边长的一半。必须为大于零的整数\n默认值： 960\n\n短边扩大模式可能提高小分辨率图片的准确度。一般用不着')
                Widget.comboboxFrame(
                    fLim, '缩放预处理：', 'ocrLimitMode', self.lockWidget, 14).pack(side='left')
                tk.Label(fLim, text='至').pack(side='left')
                wid = tk.Entry(
                    fLim, width=9, textvariable=Config.getTK('ocrLimitSize'))
                wid.pack(side='left')
                self.lockWidget.append(wid)
                tk.Label(fLim, text='像素').pack(side='left')
                # 方向
                wid = ttk.Checkbutton(frameOCR, text='启用方向分类器（文字偏转90度/180度方向矫正）',
                                      variable=Config.getTK('isOcrAngle'))
                wid.pack(side='top', fill='x', pady=2, padx=5)
                self.balloon.bind(
                    wid, '当图片中的文字偏转90度或180度时，请打开该选项\n可能略微降低识别速度\n小角度偏转无需启用该选项')
                self.lockWidget.append(wid)
                # CPU
                fCpu = tk.Frame(frameOCR, padx=5)
                fCpu.pack(side='top', fill='x')
                tk.Label(fCpu, text='线程数：').pack(side='left')
                wid = tk.Entry(
                    fCpu, width=6, textvariable=Config.getTK('ocrCpuThreads'))
                wid.pack(side='left')
                self.lockWidget.append(wid)
                self.balloon.bind(
                    wid, '最好等于CPU的线程数目。必须为大于零的整数')
                wid = ttk.Checkbutton(fCpu, text='启用MKLDNN加速',
                                      variable=Config.getTK('isOcrMkldnn'))
                wid.pack(side='left', padx=40)
                self.balloon.bind(
                    wid, '大幅加快识别速度。内存占用也会增加')
                self.lockWidget.append(wid)

                # grid
                fr1 = tk.Frame(frameOCR)
                fr1.pack(side='top', fill='x', padx=5)
                if Config.get('isAdvanced'):
                    # 隐藏高级选项：额外启动参数
                    tk.Label(fr1, text='额外启动参数：').grid(
                        column=0, row=2, sticky='w')
                    wid = tk.Entry(
                        fr1, textvariable=Config.getTK('argsStr'))
                    wid.grid(column=1, row=2, sticky="nsew")
                    self.balloon.bind(
                        wid, 'OCR高级参数指令。请遵守PaddleOCR-json要求的格式。详情参考项目主页')
                    self.lockWidget.append(wid)
                    # 隐藏高级选项：引擎管理策略
                    Widget.comboboxFrame(fr1, '引擎管理策略：', 'ocrRunMode', self.lockWidget
                                         ).grid(column=0, row=6, columnspan=2, sticky='we')
                    # 隐藏高级选项：引擎启动超时
                    fInit = tk.Frame(fr1)
                    fInit.grid(column=0, row=7, columnspan=2,
                              sticky='we', pady=2)
                    self.balloon.bind(
                        fInit, '引擎启动时，超过该时限未完成初始化，判定为启动失败')
                    tk.Label(fInit, text='初始化超时判定：').pack(side='left')
                    tk.Entry(fInit, width=5, 
                             textvariable=Config.getTK('ocrInitTimeout')).pack(side='left')
                    tk.Label(fInit, text='秒').pack(side='left')
                    
                    # 隐藏高级选项：自动清理内存
                    fRam = tk.Frame(fr1)
                    fRam.grid(column=0, row=8, columnspan=2,
                              sticky='we', pady=2)
                    tk.Label(fRam, text='自动清理内存： 占用超过').pack(side='left')
                    wid = tk.Entry(
                        fRam, width=5, textvariable=Config.getTK('ocrRamMaxFootprint'))
                    wid.pack(side='left')
                    self.lockWidget.append(wid)
                    tk.Label(fRam, text='MB 或 空闲').pack(side='left')
                    wid = tk.Entry(
                        fRam, width=5, textvariable=Config.getTK('ocrRamMaxTime'))
                    wid.pack(side='left')
                    self.lockWidget.append(wid)
                    tk.Label(fRam, text='秒').pack(side='left')
                    self.balloon.bind(
                        fRam, '引擎策略为“后台常驻”时生效\n占用内存超过指定值，或指定时间内未有任务执行，则清理一次内存\n频繁清理内存会导致卡顿，影响使用体验\n建议占用不少于 1500 MB，空闲不少于 10 秒\n两个条件独立生效。填0时忽略该条件')

                frState = tk.Frame(fr1)
                frState.grid(column=0, row=10, columnspan=2, sticky='nsew')
                tk.Label(frState, text='引擎当前状态：').pack(
                    side='left')
                tk.Label(frState, textvariable=Config.getTK('ocrProcessStatus')).pack(
                    side='left')
                labStop = tk.Label(frState, text="停止",
                                   cursor='hand2', fg="red")
                labStop.pack(side='right')
                self.balloon.bind(labStop, '强制停止引擎进程')
                labStart = tk.Label(frState, text="启动",
                                    cursor='hand2', fg='blue')
                labStart.pack(side='right', padx=5)

                def engStart():
                    try:
                        OCRe.start()
                    except Exception as err:
                        tk.messagebox.showerror(
                            '遇到了亿点小问题',
                            f'引擎启动失败：{err}')
                labStart.bind(
                    '<Button-1>', lambda *e: engStart())
                labStop.bind(
                    '<Button-1>', lambda *e: OCRe.stop())

                fr1.grid_rowconfigure(1, minsize=4)
                fr1.grid_rowconfigure(3, minsize=4)
                fr1.grid_columnconfigure(1, weight=1)
            initOcrUI()

            def initAbout():  # 关于面板
                frameAbout = tk.LabelFrame(
                    self.optFrame, text='关于')
                frameAbout.pack(side='top', fill='x', ipady=2,
                                pady=LabelFramePadY, padx=4)
                tk.Label(frameAbout, image=Asset.getImgTK(
                    'umiocr64')).pack()  # 图标
                tk.Label(frameAbout, text=Umi.name, fg='gray').pack()
                tk.Label(frameAbout, text=Umi.about, fg='gray').pack()
                labelWeb = tk.Label(frameAbout, text=Umi.website, cursor='hand2',
                                    fg='deeppink')
                labelWeb.pack()  # 文字
                labelWeb.bind(  # 绑定鼠标左键点击，打开网页
                    '<Button-1>', lambda *e: webOpen(Umi.website))
            initAbout()

            def initEX():  # 额外
                fEX = tk.Frame(self.optFrame)
                fEX.pack(side='top', fill='x', padx=4)
                labelOpenFile = tk.Label(
                    fEX, text='打开设置文件', fg='gray', cursor='hand2')
                labelOpenFile.pack(side='left')
                labelOpenFile.bind(
                    '<Button-1>', lambda *e: os.startfile('Umi-OCR_config.json'))
                self.balloon.bind(labelOpenFile, 'Umi-OCR_config.json')
                wid = tk.Checkbutton(fEX, text='调试模式', fg='gray',
                                     variable=Config.getTK('isDebug'))
                self.balloon.bind(
                    wid, '调试功能，供开发者使用，立即生效：\nOCR输出额外调试信息 | 内置截图显示调试器')
                wid.pack(side='right')
                # 隐藏高级选项
                wid = tk.Checkbutton(fEX, text='高级选项', fg='gray',
                                     variable=Config.getTK('isAdvanced'))
                self.balloon.bind(
                    wid, '启用隐藏的高级选项，重启后生效')
                wid.pack(side='right', padx=10)
                # 若初始时非置顶，不显示提示，则尾部预留出空间
                if not Config.get('isWindowTop'):
                    tk.Label(self.optFrame).pack(side='top')
            initEX()

            def initOptFrameWH():  # 初始化框架的宽高
                self.updateFrameHeight()
                self.optCanvasWidth = 1  # 宽度则是随窗口大小而改变。

                def onCanvasResize(event):  # 绑定画布大小改变事件
                    cW = event.width-3  # 当前 画布宽度
                    if not cW == self.optCanvasWidth:  # 若与上次不同：
                        self.optFrame['width'] = cW  # 修改设置页 框架宽度
                        self.optCanvasWidth = cW
                self.optCanvas.bind(  # 绑定画布大小改变事件。只有画布组件前台显示时才会触发，减少性能占用
                    '<Configure>', onCanvasResize)

                def onCanvasMouseWheel(event):  # 绑定画布中滚轮滚动事件
                    self.optCanvas.yview_scroll(
                        1 if event.delta < 0 else -1, "units")
                self.optCanvas.bind_all('<MouseWheel>', onCanvasMouseWheel)
                # 为所有复选框解绑默认滚轮事件，防止误触
                self.win.unbind_class('TCombobox', '<MouseWheel>')
            initOptFrameWH()
        initTab3()

        # 解析启动参数
        flags = Parse(argv)
        if 'error' in flags:
            tk.messagebox.showerror(
                '遇到了一点小问题', flags['error'])
        # 启动托盘
        if Config.get('isTray'):
            SysTray.start()
            self.win.wm_protocol(  # 注册窗口关闭事件
                'WM_DELETE_WINDOW', self.onCloseWin)
            # ↑ 所以，当不启动托盘时，窗口的×未关联任何事件，是默认的退出软件。
            if not flags['hide']:  # 非静默模式
                self.gotoTop()  # 恢复主窗显示
        else:  # 无托盘，强制显示主窗
            self.gotoTop()
        self.win.after(1, Config.initOK)  # 标记初始化完成
        if flags['img'] or flags['clipboard'] or flags['screenshot']:  # 有初始任务
            self.win.after(10, Mission(flags))
        Notify('欢迎使用 Umi-OCR', '通知弹窗已开启')
        self.win.mainloop()

    # 加载图片 ===============================================

    def draggedImages(self, paths):  # 拖入图片
        if not self.isMsnReady():
            tk.messagebox.showwarning(
                '任务进行中', '请停止任务后，再拖入图片')
            return
        self.notebook.select(self.notebookTab[0])  # 切换到表格选项卡
        pathList = []
        for p in paths:  # byte转字符串
            pathList.append(p.decode(Config.sysEncoding,  # 根据系统编码来解码
                            errors='ignore'))
        self.addImagesList(pathList)

    def openFileWin(self):  # 打开选择文件窗
        if not self.isMsnReady():
            return
        suf = Config.get('imageSuffix')  # 许可后缀
        paths = tk.filedialog.askopenfilenames(
            title='选择图片', filetypes=[('图片', suf)])
        self.addImagesList(paths)

    def addImagesList(self, paths):  # 添加一批图片列表
        if not self.isMsnReady():
            tk.messagebox.showwarning(
                '任务进行中', '请停止任务后，再添加图片')
            return
        suf = Config.get('imageSuffix').split()  # 许可后缀列表

        def addImage(path):  # 添加一张图片。传入路径，许可后缀。
            path = path.replace("/", "\\")  # 浏览是左斜杠，拖入是右斜杠；需要统一
            if suf and os.path.splitext(path)[1].lower() not in suf:
                return  # 需要判别许可后缀 且 文件后缀不在许可内，不添加。
            # 检测是否重复
            if self.batList.isDataItem('path', path):
                return
            # 检测是否可用
            try:
                s = Image.open(path).size
            except Exception as e:
                tk.messagebox.showwarning(
                    "遇到了一点小问题", f"图片载入失败。图片地址：\n{path}\n\n错误信息：\n{e}")
                return
            # 计算路径
            p = os.path.abspath(os.path.join(path, os.pardir))  # 父文件夹
            if not Config.get("outputFilePath"):  # 初始化输出路径
                Config.set("outputFilePath", p)
            if not Config.get("outputFileName"):  # 初始化输出文件名
                n = f"[转文字]_{os.path.basename(p)}"
                Config.set("outputFileName", n)
            # 加入待处理列表
            name = os.path.basename(path)  # 带后缀的文件名
            tableInfo = (name, "", "")
            id = self.table.insert('', 'end', values=tableInfo)  # 添加到表格组件中
            dictInfo = {"name": name, "path": path, "size": s}
            self.batList.append(id, dictInfo)

        isRecursiveSearch = Config.get("isRecursiveSearch")
        for path in paths:  # 遍历拖入的所有路径
            if os.path.isdir(path):  # 若是目录
                if isRecursiveSearch:  # 需要递归子文件夹
                    for subDir, dirs, subFiles in os.walk(path):
                        for s in subFiles:
                            addImage(subDir+"\\"+s)
                else:  # 非递归，只搜索子文件夹一层
                    subFiles = os.listdir(path)  # 遍历子文件
                    for s in subFiles:
                        addImage(path+"\\"+s)  # 添加
            elif os.path.isfile(path):  # 若是文件：
                addImage(path)  # 直接添加

    # 忽略区域 ===============================================

    def openSelectArea(self):  # 打开选择区域
        if not self.isMsnReady() or not self.win.attributes('-disabled') == 0:
            return
        defaultPath = ""
        if not self.batList.isEmpty():
            defaultPath = self.batList.get(index=0)["path"]
        self.win.attributes("-disabled", 1)  # 禁用父窗口
        IgnoreAreaWin(self.closeSelectArea, defaultPath)

    def closeSelectArea(self):  # 关闭选择区域，获取选择区域数据
        self.win.attributes("-disabled", 0)  # 启用父窗口
        area = Config.get("ignoreArea")
        if not area:
            self.ignoreFrame.pack_forget()  # 隐藏忽略区域窗口
            self.ignoreBtn.pack(side='top', fill='x')  # 显示按钮
            self.updateFrameHeight()  # 刷新框架
            return
        self.ignoreLabel["text"] = f"生效分辨率：\n宽 {area['size'][0]}\n高 {area['size'][1]}"
        self.canvas.delete(tk.ALL)  # 清除画布
        scale = self.canvasHeight / area['size'][1]  # 显示缩放比例
        width = round(self.canvasHeight * (area['size'][0] / area['size'][1]))
        self.canvas['width'] = width
        areaColor = ["red", "green", "darkorange"]
        tran = 2  # 绘制偏移量
        for i in range(3):  # 绘制新图
            for a in area['area'][i]:
                x0, y0 = a[0][0]*scale+tran, a[0][1]*scale+tran,
                x1, y1 = a[1][0]*scale+tran, a[1][1]*scale+tran,
                self.canvas.create_rectangle(
                    x0, y0, x1, y1,  fill=areaColor[i])
        self.ignoreBtn.pack_forget()  # 隐藏按钮
        self.ignoreFrame.pack(side='top', fill='x')  # 显示忽略区域窗口
        self.updateFrameHeight()  # 刷新框架

    def clearArea(self):  # 清空忽略区域
        self.ignoreFrame.pack_forget()  # 隐藏忽略区域窗口
        self.ignoreBtn.pack(side='top', fill='x')  # 显示按钮
        self.updateFrameHeight()  # 刷新框架
        Config.set("ignoreArea", None)
        self.canvas.delete(tk.ALL)  # 清除画布
        self.canvas['width'] = int(self.canvasHeight * (16/9))

    # 表格操作 ===============================================

    def clearTable(self):  # 清空表格
        if not self.isMsnReady():
            return
        self.progressbar["value"] = 0
        Config.set('tipsTop1', '')
        Config.set('tipsTop2', '请导入文件')
        Config.set("outputFilePath", "")
        Config.set("outputFileName", "")
        self.batList.clear()
        chi = self.table.get_children()
        for i in chi:
            self.table.delete(i)  # 表格组件移除

    def delImgList(self):  # 图片列表中删除选中
        if not self.isMsnReady():
            return
        chi = self.table.selection()
        for i in chi:
            self.table.delete(i)
            self.batList.delete(key=i)

    def setTableItem(self, time, score, key=None, index=-1):  # 改变表中第index项的数据信息
        if not key:
            key = self.batList.indexToKey(index)
        self.table.set(key, column='time', value=time)
        self.table.set(key, column='score', value=score)

    def clearTableItem(self):  # 清空表格数据信息
        keys = self.batList.getKeys()
        for key in keys:  # 清空表格参数
            self.table.set(key, column='time', value="")
            self.table.set(key, column='score', value="")

    # 写字板操作 =============================================

    def panelOutput(self, text, position=tk.END, highlight=''):
        '''输出面板写入文字'''
        self.textOutput.insert(position, text)
        if highlight:  # 需要高亮
            if position == tk.END:  # 暂时只允许尾部插入
                self.textOutput.tag_add(  # 尾部插入要高亮前一行
                    highlight, f'end -1lines linestart', f'end -1lines lineend')
        if self.isAutoRoll.get():  # 需要自动滚动
            self.textOutput.see(position)

    def errorOutput(self, title, msg='', highlight='red'):
        '''输出错误提示'''
        Notify(title, msg)
        if not self.textOutput.get('end-2c') == '\n':  # 当前面板尾部没有换行，则加换行
            self.panelOutput('\n')
        self.panelOutput(title, highlight=highlight)  # 输出红色提示
        if msg:
            self.panelOutput('\n'+msg)
        self.panelOutput('\n')

    def panelClear(self):
        '''清空输出面板'''
        self.textOutput.delete('1.0', tk.END)

    # 窗口操作 =============================================

    def updateFrameHeight(self):  # 刷新设置页框架高度
        self.optFrame.pack_propagate(True)  # 启用框架自动宽高调整
        self.optFrame.update()  # 强制刷新
        rH = self.optFrame.winfo_height()  # 由组件撑起的 框架高度
        self.optCanvas.config(scrollregion=(0, 0, 0, rH))  # 画布内高度为框架高度
        self.optFrame.pack_propagate(False)  # 禁用框架自动宽高调整
        self.optFrame["height"] = rH  # 手动还原高度

    def gotoTop(self, isForce=False):  # 主窗置顶
        flag = Config.get('WindowTopMode')
        # 模式：静默模式
        if flag == WindowTopModeFlag.never and not isForce and Config.get('isTray'):
            self.win.attributes('-topmost', 0)
            return
        # 模式：自动弹出，或不满足静默模式要求
        if self.win.state() == 'iconic':  # 窗口最小化状态下
            self.win.state('normal')  # 恢复前台状态
        self.win.attributes('-topmost', 1)  # 设置层级最前
        geometry = self.win.geometry()  # 缓存主窗当前位置大小
        self.win.deiconify()  # 主窗获取焦点
        self.win.geometry(geometry)  # 若主窗正在贴边，获取焦点会退出贴边模式，所以重新设置位置恢复贴边
        # 未设置窗口置顶，则一段时间后取消层级最前
        if not Config.get('isWindowTop'):
            self.win.after(500, lambda: self.win.attributes('-topmost', 0))

    # 进行任务 ===============================================

    def isMsnReady(self):
        '''可以操作下一次任务时返回T'''
        return OCRe.msnFlag == MsnFlag.none

    def setRunning(self, batFlag):  # 设置运行状态。

        def setNone():
            self.btnRun['text'] = '开始任务'
            self.btnRun['state'] = 'normal'
            Config.set('tipsTop2', '已结束')
            return 'normal'

        def initing():
            self.btnRun['text'] = '停止任务'
            self.btnRun['state'] = 'normal'
            Config.set('tipsTop1', '')
            Config.set('tipsTop2', '初始化')
            self.progressbar["maximum"] = 50  # 重置进度条长度，值越小加载动画越快
            self.progressbar['mode'] = 'indeterminate'  # 进度条为来回动模式
            self.progressbar.start()  # 进度条开始加载动画
            return 'disable'

        def running():
            self.progressbar.stop()  # 进度条停止加载动画
            self.progressbar['mode'] = 'determinate'  # 进度条静止模式
            return ''

        def stopping():
            self.btnRun['text'] = '正在停止'
            self.btnRun['state'] = 'disable'
            if str(self.progressbar["mode"]) == 'indeterminate':
                self.progressbar.stop()  # 进度条停止加载动画
                self.progressbar['mode'] = 'determinate'  # 进度条静止模式
            return ''

        state = {
            MsnFlag.none: setNone,
            MsnFlag.initing: initing,
            MsnFlag.running: running,
            MsnFlag.stopping: stopping,
        }.get(batFlag, '')()
        if state:
            for w in self.lockWidget:  # 改变组件状态（禁用，启用）
                if 'widget' in w.keys() and 'stateOFnormal' in w.keys():
                    if state == 'normal':
                        w['widget']['state'] = w['stateOFnormal']  # 正常状态为特殊值
                    else:
                        w['widget']['state'] = state
                elif 'state' in w.keys():
                    w['state'] = state
        self.win.update()

    def run(self):  # 运行按钮触发
        if self.isMsnReady():  # 未在运行
            if self.batList.isEmpty():
                return
            # 初始化批量识图任务处理器
            try:
                msnBat = MsnBatch()
            except Exception as err:
                tk.messagebox.showwarning('遇到了亿点小问题', f'{err}')
                return  # 未开始运行，终止本次运行
            # 开始运行
            paths = self.batList.getItemValueList('path')
            OCRe.runMission(paths, msnBat)
        # 允许任务进行中或初始化的中途停止任务
        elif OCRe.msnFlag == MsnFlag.running or OCRe.msnFlag == MsnFlag.initing:
            OCRe.stopByMode()

    def startSingleClipboard(self):  # 开始单张识别的剪贴板任务
        try:  # 初始化快捷识图任务处理器
            msnQui = MsnQuick()
        except Exception as err:
            tk.messagebox.showwarning('遇到了亿点小问题', f'{err}')
            return  # 未开始运行，终止本次运行
        # 开始运行
        OCRe.runMission(['clipboard'], msnQui)
        self.notebook.select(self.notebookTab[1])  # 转到输出卡
        self.gotoTop()  # 主窗置顶

    def runClipboard(self, e=None):  # 识别剪贴板
        if not self.isMsnReady():  # 正在运行，不执行
            return
        clipData = Tool.getClipboardFormat()  # 读取剪贴板

        failFlag = False

        # 剪贴板中是位图（优先）
        if isinstance(clipData, int):
            self.startSingleClipboard()

        # 剪贴板中是文件列表（文件管理器中对着文件ctrl+c得到句柄）
        elif isinstance(clipData, tuple):
            # 检验文件列表中是否存在合法文件类型
            suf = Config.get('imageSuffix').split()  # 许可后缀列表
            flag = False
            for path in clipData:  # 检验文件列表中是否存在许可后缀
                if suf and os.path.splitext(path)[1].lower() in suf:
                    flag = True
                    break
            # 存在，则将文件载入主表并执行任务
            if flag:
                self.notebook.select(self.notebookTab[0])  # 转到主表卡
                self.gotoTop()  # 主窗置顶
                self.clearTable()  # 清空主表
                self.addImagesList(clipData)  # 添加到主表
                self.run()  # 开始任务任务
            else:
                failFlag = True
        else:  # 剪贴板中不是支持的格式
            failFlag = True

        if failFlag:
            self.errorOutput('剪贴板中未查询到图片信息')
            # 失败也置顶
            self.gotoTop()  # 主窗置顶
            self.notebook.select(self.notebookTab[1])  # 转到输出卡

    def runScreenshot(self):  # 进行截图
        if not self.isMsnReady() or not self.win.attributes('-disabled') == 0:
            return
        self.win.attributes("-disabled", 1)  # 禁用主窗口
        if Config.get('isScreenshotHideWindow'):  # 截图时隐藏主窗口
            self.win.state('iconic')
            self.win.after(Config.get('screenshotHideWindowWaitTime'),
                           ScreenshotCopy)  # 延迟，等待最小化完成再截屏
        else:
            ScreenshotCopy()  # 立即截屏

    def openScreenshot(self, e=None):  # 普通截图
        Config.set('isFinishSend', False)
        self.runScreenshot() 

    def openLinkageScreenshot(self, e=None):  # 联动截图
        Config.set('isFinishSend', True)
        self.runScreenshot() 

    def closeScreenshot(self, flag, errMsg=None):  # 关闭截图窗口，返回T表示已复制到剪贴板
        self.win.attributes("-disabled", 0)  # 启用父窗口
        if errMsg:
            self.errorOutput('截图失败', errMsg)
        if not flag and self.win.state() == 'normal':  # 截图不成功，但窗口非最小化
            self.gotoTop()  # 主窗置顶
        elif flag:  # 成功
            # self.win.after(50, self.runClipboard)
            self.startSingleClipboard()  # 剪贴板识图

    def onCloseWin(self):  # 关闭窗口事件
        if Config.get('isBackground'):
            self.win.withdraw()  # 隐藏窗口
        else:
            self.onClose()  # 直接关闭

    def onClose(self):  # 关闭软件
        OCRe.stop()  # 强制关闭引擎进程，加快子线程结束
        if OCRe.engFlag == EngFlag.none and OCRe.msnFlag == MsnFlag.none:  # 未在运行
            self.exit()
        else:
            self.win.after(50, self.waitClose)  # 等待关闭，50ms轮询一次是否已结束子线程

    def waitClose(self):  # 等待线程关闭后销毁窗口
        Log.info(f'关闭中，等待 {OCRe.engFlag} | {OCRe.msnFlag}')
        if OCRe.engFlag == EngFlag.none and OCRe.msnFlag == MsnFlag.none:  # 未在运行
            self.exit()
        else:
            self.win.after(50, self.waitClose)  # 等待关闭，50ms轮询一次是否已结束子进程

    def exit(self):
        SysTray.stop()  # 关闭托盘。这个函数里有判断，不会造成无限递归。
        # 等待一段时间，保证托盘线程关闭，图标从系统注销
        # 然后强制终止主进程，防止引擎子线程苟且偷生
        self.win.after(100, lambda: os._exit(0))

    def showTips(self, tipsText):  # 显示提示
        if not self.isMsnReady():
            tk.messagebox.showwarning(
                '任务进行中', '请停止任务后，再打开软件说明')
            return
        self.notebook.select(self.notebookTab[1])  # 切换到输出选项卡
        outputNow = self.textOutput.get("1.0", tk.END)
        if outputNow and not outputNow == "\n":  # 输出面板内容存在，且不是单换行（初始状态）
            if not tkinter.messagebox.askokcancel('提示', '将清空输出面板。要继续吗？'):
                return
            self.panelClear()
        self.textOutput.insert(tk.END, tipsText)

# 全角空格：【　】



from utils.logger import GetLog
from utils.config import Config, RunModeFlag
from ocr.api_ppocr_json import OcrAPI
from ocr.engine_ram_optimization import OcrEngRam

import os
import time
import asyncio
import threading
from operator import eq
from enum import Enum

Log = GetLog()


class EngFlag(Enum):
    '''引擎运行状态标志'''
    none = 0  # 不在运行
    initing = 1  # 正在启动
    waiting = 2  # 待命
    running = 3  # 工作中


class MsnFlag(Enum):
    '''批量任务状态标志'''
    none = 0  # 不在运行
    initing = 1  # 正在启动
    running = 2  # 工作中
    stopping = 3  # 停止中


class OcrEngine:
    '''OCR引擎，含各种操作的方法'''

    def __init__(self):
        # self.__initVar() # 不能使用__initVar，不能调用self.setEngFlag()，因为不能保证主tk已经启动事件循环
        self.__ocrInfo = ()  # 记录之前的OCR参数
        self.__ramTips = ''  # 内存占用提示
        self.__runMissionLoop = None  # 批量识别的事件循环
        self.ocr = None  # OCR API对象
        self.winSetRunning = None
        self.engFlag = EngFlag.none
        self.msnFlag = MsnFlag.none
        OcrEngRam.init(self.restart, self.getEngFlag, EngFlag)  # 内存优化·初始化，传入接口

    def __initVar(self):
        self.__ocrInfo = ()  # 记录之前的OCR参数
        self.__ramTips = ''  # 内存占用提示
        self.ocr = None  # OCR API对象
        # self.msnFlag = MsnFlag.none  # 任务状态不能在这里改，可能引擎已经关了，任务线程还在继续

    def __setEngFlag(self, engFlag):
        '''更新引擎状态并向主窗口通知'''
        self.engFlag = engFlag
        if self.ocr and Config.get('isDebug'):
            if engFlag == EngFlag.waiting:  # 刷新内存占用
                self.__ramTips = f'（内存：{self.ocr.getRam()}MB）'
        msg = {
            EngFlag.none:  '已关闭',
            EngFlag.initing:  '正在启动',
            EngFlag.waiting:  f'待命{self.__ramTips}',
            EngFlag.running:  f'工作{self.__ramTips}',
        }.get(engFlag, f'未知（{engFlag}）')
        isTkUpdate = False
        if engFlag == EngFlag.initing:  # 启动中，刷新一下UI
            isTkUpdate = True
        Config.set('ocrProcessStatus', msg, isTkUpdate)  # 设置
        # Log.info(f'引擎 ⇒ {engFlag}')

    def getEngFlag(self):
        return self.engFlag

    def __setMsnFlag(self, msnFlag):
        '''更新任务状态并向主窗口通知'''
        self.msnFlag = msnFlag
        if self.winSetRunning:
            self.winSetRunning(msnFlag)
        # Log.info(f'任务 ⇒ {msnFlag}')

    @staticmethod
    def __tryFunc(func, *e):
        '''尝试执行func'''
        if func:
            try:
                func(*e)
            except Exception as e:
                errMsg = f'调用函数 {str(func)} 异常： {e}'
                Log.error(errMsg)
                Config.main.panelOutput(errMsg+'\n')

    def start(self):
        '''启动引擎。若引擎已启动，且参数有更新，则重启。'''
        if self.engFlag == EngFlag.initing:  # 正在初始化中，严禁重复初始化
            return
        # 检查引擎路径
        ocrToolPath = Config.get('ocrToolPath')
        if not os.path.isfile(ocrToolPath):
            raise Exception(
                f'未在以下路径找到引擎组件\n【{ocrToolPath}】\n\n请将引擎组件【PaddleOCR-json】文件夹放置于指定路径！')
        # 获取静态参数
        ang = ' -cls=1 -use_angle_cls=1' if Config.get('isOcrAngle') else ''
        limit = f" -limit_type={Config.get('ocrLimitMode').get(Config.get('ocrLimitModeName'),'min')} -limit_side_len={Config.get('ocrLimitSize')}"
        staticArgs = f"{ang}{limit}\
 -cpu_threads={Config.get('ocrCpuThreads')}\
 -enable_mkldnn={Config.get('isOcrMkldnn')}\
 {Config.get('argsStr')}"  # 静态启动参数字符串。注意每个参数前面的空格
        # 整合最新OCR参数
        info = (
            ocrToolPath,  # 识别器路径
            Config.get('ocrConfig')[Config.get(
                'ocrConfigName')]['path'],  # 配置文件路径
            staticArgs,  # 启动参数
        )
        isUpdate = not eq(info, self.__ocrInfo)  # 检查是否有变化

        if self.ocr:  # OCR进程已启动
            if not isUpdate:  # 无变化则放假
                return
            self.stop(True)  # 有变化则先停止OCR进程再启动。传入T表示是在重启，无需中断任务。

        self.__ocrInfo = info  # 记录参数。必须在stop()之后，以免被覆盖。
        try:
            Log.info(f'启动引擎，参数：{info}')
            self.__setEngFlag(EngFlag.initing)  # 通知启动中
            self.ocr = OcrAPI(*self.__ocrInfo, initTimeout=Config.get('ocrInitTimeout'))  # 启动引擎
            # 检查启动引擎这段时间里，引擎有没有被叫停
            if not self.engFlag == EngFlag.initing:  # 状态被改变过了
                Log.info(f'初始化后，引擎被叫停！{self.engFlag}')
                self.stop()
                return
            self.__setEngFlag(EngFlag.waiting)  # 通知待命
        except Exception as e:
            self.stop()
            raise

    def stop(self, isRestart=False):
        '''立刻终止引擎。isRE为T时表示这是在重启，无需中断任务。'''
        if (self.msnFlag == MsnFlag.initing or self.msnFlag == MsnFlag.running)\
                and not self.engFlag == EngFlag.none and not isRestart:
            Log.info(f'引擎stop，停止任务！')
            self.__setMsnFlag(MsnFlag.stopping)  # 设任务需要停止
        if hasattr(self.ocr, 'stop'):
            self.ocr.stop()
        del self.ocr
        self.ocr = None
        self.__setEngFlag(EngFlag.none)  # 通知关闭
        self.__initVar()

    def stopByMode(self):
        '''根据配置模式决定是否停止引擎'''
        if self.msnFlag == MsnFlag.initing or self.msnFlag == MsnFlag.running\
                and not self.engFlag == EngFlag.none:
            self.__setMsnFlag(MsnFlag.stopping)  # 设任务需要停止
        n = Config.get('ocrRunModeName')
        modeDict = Config.get('ocrRunMode')
        if n in modeDict.keys():
            mode = modeDict[n]
            if mode == RunModeFlag.short:  # 按需关闭
                self.stop()

    def restart(self):
        '''重启引擎，释放内存'''
        self.stop(True)
        self.start()

    def run(self, path):
        '''执行单张图片识别，输入路径，返回字典'''
        if not self.ocr:
            self.__setEngFlag(EngFlag.none)  # 通知关闭
            return {'code': 404, 'data': f'引擎未在运行'}
        OcrEngRam.runBefore(ram=self.ocr.getRam())  # 内存优化·前段
        self.__setEngFlag(EngFlag.running)  # 通知工作
        data = self.ocr.run(path)
        # 有可能因为提早停止任务或关闭软件，引擎被关闭，OCR.run提前出结果
        # 此时 engFlag 已经被主线程设为 none，如果再设waiting可能导致bug
        # 所以检测一下是否还是正常的状态 running ，没问题才通知待命
        if self.engFlag == EngFlag.running:
            self.__setEngFlag(EngFlag.waiting)  # 通知待命
        OcrEngRam.runAfter()  # 内存优化·后段
        return data

    def runMission(self, paths, msn):
        '''批量识别多张图片，异步。若引擎未启动，则自动启动。\n
        paths: 路径\n
        msn:   任务器对象，Msn的派生类，必须含有 onStart|onGet|onStop|onError 四个方法'''
        if not self.msnFlag == MsnFlag.none:  # 正在运行
            Log.error(f'已有任务未结束就开始了下一轮任务')
            raise Exception('已有任务未结束')

        self.winSetRunning = Config.main.setRunning  # 设置运行状态接口
        self.__setMsnFlag(MsnFlag.initing)  # 设任务初始化

        def runLoop():  # 启动事件循环
            asyncio.set_event_loop(self.__runMissionLoop)
            self.__runMissionLoop.run_forever()

        # 在当前线程下创建事件循环
        self.__runMissionLoop = asyncio.new_event_loop()
        # 开启新的线程，在新线程中启动事件循环
        threading.Thread(target=runLoop).start()
        # 在新线程中事件循环不断游走执行
        asyncio.run_coroutine_threadsafe(self.__runMission(
            paths, msn
        ), self.__runMissionLoop)

    async def __runMission(self, paths, msn):
        '''新线程中批量识图。在这个线程里更新UI是安全的。'''

        num = {
            'all': len(paths),  # 全部数量
            'now': 1,  # 当前处理序号
            'index': 0,  # 当前下标
            'succ': 0,  # 成功个数
            'err': 0,  # 失败个数
            'exist': 0,  # 成功里面有文字的个数
            'none': 0,  # 成功里面无文字的个数
            'time': 0,  # 执行至今的总时间
            'timeNow': 0,  # 这一轮的耗时
        }

        def close():  # 停止
            try:
                self.__runMissionLoop.stop()  # 关闭异步事件循环
            except Exception as e:
                Log.error(f'任务线程 关闭任务事件循环失败： {e}')
            self.stopByMode()  # 按需关闭OCR进程
            self.__tryFunc(msn.onStop, num)
            self.__setMsnFlag(MsnFlag.none)  # 设任务停止
            Log.info(f'任务close！')

        # 启动OCR引擎，批量任务初始化 =========================
        try:
            self.start()  # 启动或刷新引擎
        except Exception as e:
            Log.error(f'批量任务启动引擎失败：{e}')
            self.__tryFunc(msn.onError, num, f'无法启动引擎：{e}')
            close()
            return
        timeStart = time.time()  # 启动时间
        timeLast = timeStart  # 上一轮结束时间

        # 检查启动引擎这段时间里，任务有没有被叫停 =========================
        if self.msnFlag == MsnFlag.stopping:  # 需要停止
            close()
            return
        # 主窗UI和任务处理器初始化 =========================
        self.__setMsnFlag(MsnFlag.running)  # 设任务运行
        self.__tryFunc(msn.onStart, num)

        # 正式开始任务 =========================
        for path in paths:
            if self.msnFlag == MsnFlag.stopping:  # 需要停止
                close()
                return
            isAddErr = False
            try:
                data = self.run(path)  # 调用图片识别
                # 刷新时间
                timeNow = time.time()  # 本轮结束时间
                num['time'] = timeNow-timeStart
                num['timeNow'] = timeNow-timeLast
                timeLast = timeNow
                # 刷新量
                if data['code'] == 100:
                    num['succ'] += 1
                    num['exist'] += 1
                elif data['code'] == 101:
                    num['succ'] += 1
                    num['none'] += 1
                else:
                    num['err'] += 1
                    isAddErr = True
                    # 若设置了进程按需关闭，中途停止任务会导致进程kill，本次识别失败
                    # 若设置了进程后台常驻，中途停止任务会让本次识别完再停止任务
                    # 这些都是正常的（设计中的）
                    # 但是，引擎进程意外关闭导致图片识别失败，则是不正常的；所以不检测engFlag
                    if self.msnFlag == MsnFlag.stopping:  # 失败由强制停止引擎导致引起
                        data['data'] = '这是正常情况。中途停止任务、关闭引擎，导致本张图片未识别完。'
                # 调用取得事件
                self.__tryFunc(msn.onGet, num, data)
            except Exception as e:
                Log.error(f'任务线程 OCR失败： {e}')
                if not isAddErr:
                    num['err'] += 1
                continue
            finally:
                num['now'] += 1
                num['index'] += 1

        close()


OCRe = OcrEngine()  # 引擎单例
